package editor;
import gui.CanvasWindow;
import gui.ImageArray;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;
import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.KeyStroke;
import javax.swing.LookAndFeel;
import javax.swing.Timer;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.colorchooser.AbstractColorChooserPanel;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.Element;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledEditorKit;
import javax.swing.text.html.CSS;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLDocument.BlockElement;
import javax.swing.text.html.HTMLDocument.Iterator;
import javax.swing.text.html.StyleSheet;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoManager;

import rtf.HTML2RTF;
import utils.ASyntax;
import utils.Data;
import utils.DataUsers;
import utils.Global;
import virtual.TecladoVirtual1;
import virtual.TecladoVirtual2;

/**
 * Builds a maximised non-resizable window with two toolbars: a horizontal one at the top for text editing,
 * and a vertical one on the left-hand side for real-time accessibility features. The rest of the space is 
 * taken up by a text area. Save some exceptions, the visibility of the toolbar buttons is configurable. 
 * The text itself is rendered and stored using HTML 3.2. By default, the editor opens with the most 
 * recently modified document and automatically saves it at regular intervals.  
 * @author hqcasanova
 * 
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 */
public class Editor extends CanvasWindow implements ActionListener,
 													 KeyListener,
 													 CaretListener,
 													 UndoableEditListener,
													 WindowListener {
	
	//Interactive elements
	private JButton newDoc, switchUsr, credits, //permanently visible buttons
	                printDoc;					//print document button  
	private JTextPane textArea;			        //style-enabled editable area 
	private JTextPane textAreaCopy;				//copy of currently active editable area
	private JLabel avatar;						//current user's avatar image
	private JPopupMenu vKeypadMenu;				//popup menu for available virtual keypads
        
    //Secondary windows and dialogues
    private final CanvasWindow PBDQkey;			//window with colour key for the PBDQ feature
	private final Login login;					//login dialogue
	private final FieldConfirm tableConfig;		//table configuration dialogue
	private final JFileChooser imageChooser;	//dialogue for image file selection.
	private final JDialog colorChooser;			//dialogue for colour selection
	private final JColorChooser colorPane;		//colour chooser pane to be displayed in colour dialogue
		
	//Editing infrastructure
	private ExtendedHTMLKit editorKit;		    //support kit for HTML formating
	private HTMLDocument document;				//data model underlying the typed text
	private StyleSheet globalStyles;  	        //CSS styles affecting the whole document
	private UndoManager undo;		            //managing facility for supporting undo actions
	private HTML2RTF RTFConverter; 				//HTML parser for conversion to RTF code and saving
    		
	//Others
	private HashMap<String, Button> buttons;    //toolbar buttons
	private FileNameExtensionFilter imgFilter;	//file filter for the image file selection dialogue
    private Timer refresh,						//used to force refresh of text area
                  saveTimer;				    //auto-save timer
	private Icon pasteIcon, copyIcon;	    	//icon images for copy/paste button	
	private boolean bCopy, 				    	//has the copy button been pressed?
	                bBuildingList,				//is a list being built?
	                bNewListItem,				//ENTER key been pressed at empty list item? => new item
	                bPictosEnabled,			    //pictograms enabled in current user's profile?
	                bPictosOn,			    	//pictogram feature on?	
	                bPBDQOn,			    	//PBDQ feature on?
	                bVKeypadOn,					//virtual keypad feature on?
					bGrabFocus,					//transfer focus to text area after editing action?
	                bPurgeKeyPress;				//cancel any pending event after key press?
	private int iUsrId, 					    //current user id
	            iCurrBgColour, 			        //current background colour index
	            iAutoPictoID;					//auto-incrementing pictogram ID;
	private String sUsrName,				    //current user's name
 	               sProfilePath,				//current user's profile path
 	               sDocName,					//current document's name
	               sDocId; 				        //current document's ID                   
     	
 	//background colours for the text area
 	private static final Color[] bgColours = {Color.decode ("#FFFFFF"),
 										      Color.decode ("#000000"),
                                              Color.decode ("#FFFF80")};	    
 	
 	//background-to-foreground colour mapping
 	@SuppressWarnings("serial")
	private static final HashMap<Color, Color> fgColours = new HashMap<Color, Color>() {{
 								put (bgColours[0], Color.BLACK);
 								put (bgColours[1], Color.WHITE);
 								put (bgColours[2], Color.BLACK); }};
 	
 	//map of bacground colour index to PBDQ font colour
 	private static final Color[] colorP = {Color.decode ("#0000C0"), Color.decode ("#40FFFF"), Color.decode ("#4040FF")}, 
 					             colorB = {Color.decode ("#606060"), Color.decode ("#C0C0C0"), Color.decode ("#606060")}, 
 					             colorD = {Color.decode ("#006D00"), Color.decode ("#40FF40"), Color.decode ("#006D00")}, 
 					             colorQ = {Color.decode ("#A300A3"), Color.decode ("#FF00FF"), Color.decode ("#A300A3")};
 								
 	//map of character to PBDQ background-dependant colour
 	private final HashMap<Character, Color[]> mapPBDQ; 
 	
 	//HTML representation of an empty paragraph
 	private static final String sEmptyParagraph = "<p style=\"margin-top: 0\"></p>";
 	
 	//width and height of inline pictograms.
 	private static final int iPictoSize = 2 * Global.ICON_SIZE;
 		
 	/*
 	 * Java's support for style-enabled HTML editing can be considered to be organised 
 	 * around 3 classes, each representing a layer or view of the document. Namely, from top to bottom:
 	 *  
 	 * JTextPane (extends JEditorPane)				= 	Caret view (unit: offset from start of document)
 	 * -------------------------------------------------------------------------------------------------
 	 * HTMLDocument (extends DefaultStyledDocument)	=	Data model view (unit: element)
 	 * -------------------------------------------------------------------------------------------------
 	 * HTMLEditorKit (extends StyledEditorKit)		= 	HTML view 
 	 * 
 	 * Text insertions should be performed at the data model level whenever possible. This practice helps
 	 * prevent accidental overwrite of the document's header or any other part not represented in the 
 	 * "visible-version" of it.
 	 * 
 	 * In HTMLEditorKit, line breaks are coded as paragraphs. Generally, the correct rendering of 
 	 * an element seems to be guaranteed only if it is enclosed in a paragraph. It follows that there 
 	 * should always be at least one paragraph in the document, even if its empty. This is exactly the 
 	 * implemented policy in the present editor.
 	 * 
 	 * In HTML, continuous whitespace sequences coalesce to a single whitespace. Hence, nbsp entities will
 	 * have to be used instead of regular whitespaces to avoid that happening (especially when the HTML parser kicks in). 
 	 * However, that prevents proper line-wrapping since all lines are effectively continuous strings of HTML (with no 
 	 * whitespaces). This is why a customised HTMLEditorKit class is used (overwriting exactly the methods dealing with 
 	 * line-wrapping). 
 	 */
 	
	/**
	 * Customised constructor. Draws the window, determines button visibility, sets up the 
	 * document and its timed saving, sets up the virtual keypad menu and shows the dialogue for
	 * creating a new document if no previous one exists. Guarantees the saving of the opened document 
	 * in case of an abrupt exit. 
	 * @param sTitle Editing window's title
	 * @param iFontSize Font size set as the default
	 * @param sLookFeel Name of the currently installed look and feel
	 * @param progress Progress bar object for the splash screen
	 * @param splash Window serving as splash screen
	 */
	@SuppressWarnings("serial")
	public Editor (String sTitle, int iFontSize, String sLookFeel, JProgressBar progress,
			       CanvasWindow splash) {
		super (sTitle, null, null, null, false, JFrame.DO_NOTHING_ON_CLOSE, Global.FAVICON_ED, false);
		
	    //Interactive elements
		JScrollPane scroll;       		//scrollable view of the text area
		JPanel topPanel,				//section above the text area 
		       centrePanel,				//section comprising side toolbar and text area
		       pagedArea;				//contents of the scroll pane
		final Credits creditDiag;		//dialogue with credits
					
		//Others
		final AbstractColorChooserPanel[] chooserPanels;  //default chooser panels in colour chooser
		final String sIconPath;			//path for button icons		             
		final String[] tableFormats = {Global.TABLE_REGEX, Global.TABLE_REGEX};
		Font newFont;					//font installed by the editor to the JVM (system-wide guarantee)
		ImageArray iconImages;

		progress.setValue (1);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Sets user and document data
		iUsrId = 0;
		sUsrName = "";
		sDocId = "";
		sDocName = "";
		sProfilePath = "";
		sIconPath = Global.IMG_FOLDER + Global.SEPARADOR_SYS;
		
		//Initialise other fields
	    iCurrBgColour = 0;
	    iAutoPictoID = 0;
		
		//Sets up all needed elements for style-enabled editing
		textArea = new JTextPane();
		if (sLookFeel.equals ("Nimbus") ) { //ugly fix so background honoured in Nimbus
		   textArea.setUI (new javax.swing.plaf.basic.BasicTextPaneUI());
		}
		editorKit = new ExtendedHTMLKit();
		textArea.setEditorKit (editorKit);
		textArea.setContentType ("text/html;charset=UTF-8");
	    textArea.addKeyListener (this);
	    textArea.setBackground (bgColours[iCurrBgColour]);
	    textAreaCopy = new JTextPane ();
	    textAreaCopy.setDoubleBuffered (false);
	    textAreaCopy.setEditorKit (new ExtendedHTMLKit());
		textAreaCopy.setContentType ("text/html;charset=UTF-8"); 
		
		//Adjust text area's width to page size
		//textArea.setPreferredSize (new Dimension (Global.PAGE_WIDTH, Short.MAX_VALUE));
	    textArea.setBorder (BorderFactory.createCompoundBorder (
	    		                  BorderFactory.createLineBorder (bgColours[iCurrBgColour], Global.PADDING_PIX),
	    		                  BorderFactory.createLineBorder (Global.DEFAULT_BORDER, 1)));
	    /*pagedArea = new JPanel();
		pagedArea.setBackground (Color.GRAY);
		pagedArea.setBorder (BorderFactory.createEmptyBorder (10 * Global.BUTTON_EDGE, 0, 10 * Global.BUTTON_EDGE, 0));
		pagedArea.add (textArea);
		scroll.setViewportView (pagedArea);*/
		
		progress.setValue (2);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Sets up window's centre panel
		centrePanel = new JPanel (new BorderLayout (5, 5));
		scroll = new JScrollPane (textArea);
		centrePanel.add (scroll, BorderLayout.CENTER);
		centrePanel.add (new JLabel(), BorderLayout.WEST);
		
		//Sets maximum possible size for window and reports it to the system 
		maximise();		
		
		//Adds the text area, taking up the whole window.
		populate (centrePanel);
        getWindow().addWindowListener (this);
		
        //Displays a stripped-down version of the editor (without buttons).
		reset();
		getWindow().setVisible (true);
		progress.setValue (3);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Shows busy cursor
		busy (true);
		
		//Inserts the welcome banner at the top of the text area
		iconImages = new ImageArray (Global.BANNER, sIconPath, Global.PICT_EXT);
	    iconImages.resize (textArea.getWidth() - (2 * Global.PADDING_PIX), -1);
	    textArea.insertIcon (iconImages.toIcon());
	    textArea.paintImmediately (textArea.getVisibleRect());
	        	    
	    //Sets up the RTF converter
		RTFConverter = new HTML2RTF();
		progress.setValue (4);
		progress.paintImmediately (progress.getVisibleRect());
	    
		//Installs default editor font
	    try {
	    	newFont = Font.createFont (Font.TRUETYPE_FONT, new File (Global.FONT_FAMILY +".ttf"));
	    	Global.GRAPHICS.registerFont (newFont);	
	    	progress.setValue (5); 
	    	progress.paintImmediately (progress.getVisibleRect());
	    //If anything goes wrong, shows an error message.
	    } catch (Exception e) {
	    	    newFont = null;
	    	    JOptionPane.showMessageDialog
                            (getWindow(), Global.LANG.getString ("fuenteMensaje") +Global.FONT_FAMILY+ ".ttf", 
                             Global.LANG.getString ("fuenteTitulo"), JOptionPane.INFORMATION_MESSAGE);	    	    
	    }
	    
	    //Set the application-wide default language and open the corresponding database.
	    selectLanguage (getWindow(), splash);
	    Global.DB_HANDLER = new ConnectDB (Global.INSTALL_FOLDER + Global.SEPARADOR_SYS + 
	    		                           Global.DB_FILE);
	    progress.setValue (6);
	    progress.paintImmediately (progress.getVisibleRect());
	    
		//Initialise flags
		bBuildingList = false;
		bNewListItem = false;
		bPurgeKeyPress = false;
		bGrabFocus = false;
		
		//Builds PBDQ colour map
		mapPBDQ = new HashMap<Character, Color[]>(); 
	    mapPBDQ.put ('p', colorP);
		mapPBDQ.put ('b', colorB);
		mapPBDQ.put ('d', colorD);
		mapPBDQ.put ('q', colorQ);
		
		//Sets the window with PBDQ's colour key
        PBDQkey = new CanvasWindow (Global.LANG.getString("pbdqTitulo"), null, null, 
				                    Color.WHITE, false, JFrame.DISPOSE_ON_CLOSE, 
				                    Global.FAVICON_ED, false);
		iconImages = new ImageArray ("leyendaPBDQ", sIconPath, Global.PICT_EXT);
		iconImages.resize (Global.FRAME_WIDTH, Global.FRAME_HEIGHT);
		PBDQkey.populate (new JLabel (iconImages.toIcon()));
		progress.setValue (7);
		progress.paintImmediately (progress.getVisibleRect());
	    
	    //Sets up the create/load a new document
		newDoc = CanvasWindow.toSwing (Global.ICON_SIZE, Global.ICON_SIZE, sIconPath, Global.OPENDOC, 
				                   Global.PICT_EXT, Global.LANG.getString ("editorBotonNuevo"), 
				                   this);
		progress.setValue (8);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Sets up the credits dialogue
		creditDiag = new Credits (Global.LANG.getString("creditTitulo") + Global.VERSION, getWindow(),
				                  Global.FAVICON_ED, Global.FRAME_WIDTH, sIconPath, Global.TEMP_FOLDER, "ed");
		progress.setValue (9);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Sets up the credits button
		credits = CanvasWindow.toSwing (Global.ICON_SIZE, Global.ICON_SIZE, sIconPath, Global.CREDITS_IMG, 
				                    Global.PICT_EXT, Global.LANG.getString ("editorBotonCreditos"), 
				                    new ActionListener() {
				  		                 public void actionPerformed (ActionEvent e) {creditDiag.display();}
				       	           });
		credits.setAlignmentX (Component.CENTER_ALIGNMENT);
		progress.setValue (10);
		progress.paintImmediately (progress.getVisibleRect());
		
		/*Sets up the print document button
		final Paper paper = new Paper();
		paper.setSize (Global.PAGE_WIDTH, Global.PAGE_HEIGHT);
		printDoc = Button.buttonise (Global.ICON_SIZE, Global.ICON_SIZE, sIconPath, Global.PRINT, 
                                     Global.PICT_EXT, Global.LANG.getString ("editorBotonImprimir"), 
                                     new ActionListener() {
		                 			     public void actionPerformed (ActionEvent e) {
		                 			    	 try {		                 			    		 
		                 			    		 //Set up print job: page size, margins, job name
		                 			    		 HashPrintRequestAttributeSet attr = new HashPrintRequestAttributeSet();
		                 			    		 MediaSize pageDim = MediaSize.getMediaSizeForName (MediaSizeName.ISO_A4);
		                 			    		 attr.add (MediaSizeName.ISO_A4);
		                 			    		 attr.add (new MediaPrintableArea (Global.PADDING_INCH, Global.PADDING_INCH, 
		                 			    				                           pageDim.getX (MediaSize.INCH) - 2 * Global.PADDING_INCH, 
		                 			    				                           pageDim.getY (MediaSize.INCH) - 2 * Global.PADDING_INCH, 
		                 			    				                           MediaPrintableArea.INCH));
		                 			    		 attr.add (new JobName (sUsrName +": "+ sDocName, Global.LANG.getLocale()));
		                 			    		 
		                 			    		 //Get hidden text area ready for print: update and convert font size to points
		                 			    		 saveDoc();
		                 			    		 HTMLDocument docModel = (HTMLDocument) textAreaCopy.getDocument();
		                 			    		 docModel.getStyleSheet().addRule ("body {font-size: " +Global.FONT_SIZE+ "pt; font-family: " +Global.FONT_FAMILY+ "}");
		                 			    		 textAreaCopy.setText (textAreaCopy.getText());
		                 			    		 textAreaCopy.repaint();
		                 			    		 
		                 			    		 //Show print dialogue => send job to printer => show progress
		                 			    		 textAreaCopy.print (null, null, true, 
														             PrinterJob.getPrinterJob().getPrintService(), 
														             attr, true);
		                 			    		 
		                 			    		 //Restore properties of hidden text area 
		                 			    		 docModel.getStyleSheet().addRule ("body {font-size: " +Global.FONT_SIZE+ "px; font-family: " +Global.FONT_FAMILY+ "}");
		                 			    		 textAreaCopy.setText (textAreaCopy.getText());	
		                 			    		 textAreaCopy.repaint();
											 } catch (PrinterException ePrint) {
												     System.out.println ("An error has occurred while printing: " +ePrint);
													 ePrint.printStackTrace();
											 }       		                 			    	  
		                 			      }
   	           						});
		printDoc.setAlignmentX (Component.CENTER_ALIGNMENT);*/
        progress.setValue (11);
        progress.paintImmediately (progress.getVisibleRect());
		
		avatar = new JLabel ();
								
		//Sets alternating icons for the copy/paste button
		iconImages = new ImageArray ("pegar", sIconPath, Global.PICT_EXT);
		iconImages.resize (Global.ICON_SIZE, Global.ICON_SIZE);
		pasteIcon = iconImages.toIcon();
		iconImages = new ImageArray ("copiarOPegar", sIconPath, Global.PICT_EXT);
		iconImages.resize (Global.ICON_SIZE, Global.ICON_SIZE);
		copyIcon = iconImages.toIcon();
		progress.setValue (12);
		progress.paintImmediately (progress.getVisibleRect());
					
		//Sets up the table insert dialogue
		final String[] tableFields = {Global.LANG.getString("tablaFilTexto"), Global.LANG.getString("tablaColTexto")};
		tableConfig = new FieldConfirm (Global.LANG.getString("tablaTitulo"), tableFields, tableFormats);	
		progress.setValue (13);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Sets up the image file selection dialogue 
		imageChooser = new JFileChooser (System.getProperty ("user.home"));
	    imgFilter = new FileNameExtensionFilter (Global.LANG.getString("nombreFiltro"), "jpg", "gif", "png");
	    	        
	    //Sets up the dialogue for colour selection, leaving only the swatches panel
	    UIManager.put ("ColorChooser.resetText", Global.LANG.getString ("botonReset"));
	    UIManager.put ("ColorChooser.okText", Global.LANG.getString ("botonAceptar"));
	    UIManager.put ("ColorChooser.cancelText", Global.LANG.getString ("botonCancelar"));
	    UIManager.put ("ColorChooser.sampleText", Global.LANG.getString ("colorPrueba"));
	    UIManager.put ("ColorChooser.swatchesRecentText", Global.LANG.getString ("colorRecientes"));
	    colorPane = new JColorChooser (fgColours.get(bgColours[iCurrBgColour]));
		chooserPanels = colorPane.getChooserPanels();
  	    for (int i=0; i<chooserPanels.length; i++) {
		    String clsName = chooserPanels[i].getClass().getName();
		    if (!clsName.equals("javax.swing.colorchooser.DefaultSwatchChooserPanel")) {
		       colorPane.removeChooserPanel (chooserPanels[i]);
		    } 
		}
		colorChooser = JColorChooser.createDialog (getWindow(), Global.LANG.getString("colorTitulo"), true, 
				                    			   colorPane, null, null);
		progress.setValue (14);
	    progress.paintImmediately (progress.getVisibleRect());
		
	    //Sets up the login dialogue
	    login = new Login (Global.LANG.getString ("loginTitulo"), 
				(int) (Global.SCREEN_DIM.height / 10.5), (int) (Global.SCREEN_DIM.height / 105), 
				(int) (Global.SCREEN_DIM.height / 10.5), (int) (Global.SCREEN_DIM.height / 10.5), this);
	    progress.setValue (15);    	  
	    progress.paintImmediately (progress.getVisibleRect());
	    
	    //Sets up the user switch button
	    switchUsr = CanvasWindow.toSwing (Global.ICON_SIZE, Global.ICON_SIZE, sIconPath, Global.USER_SWITCH, 
	    		                      Global.PICT_EXT, Global.LANG.getString ("nuevoBotonCambioUsr"), 
	    		                      new ActionListener() {
	    		              		       public void actionPerformed (ActionEvent e) {login.display();}
	    		                   	 });
	    progress.setValue (16);
		progress.paintImmediately (progress.getVisibleRect());
		
		//Sets up the top panel and adds the buttons at both corners to it
		topPanel = new JPanel ();
		topPanel.setLayout (new BoxLayout (topPanel, BoxLayout.X_AXIS));
		topPanel.add (newDoc); 
		topPanel.add (Box.createHorizontalGlue());
		topPanel.add (new JLabel());
		topPanel.add (Box.createHorizontalGlue());
		topPanel.add (switchUsr);
		topPanel.add (Box.createRigidArea(new Dimension (Global.BUTTON_EDGE, 0)));
		topPanel.add (avatar);
		changeTopPanel (topPanel);
		getTopPanel().setVisible(false);
		progress.setValue (17);
		progress.paintImmediately (progress.getVisibleRect());
				    
		//Translates the tab character into a series of HTML non-breaking spaces 
	    textArea.getInputMap().put (KeyStroke.getKeyStroke ("TAB"), "tab" );  
	    textArea.getActionMap().put ("tab", new AbstractAction() {  
	      	     public void actionPerformed (ActionEvent e) {  
	      	    	 replaceString (Global.TAB, textArea.getSelectionStart(), textArea.getSelectionEnd(), 
	      	    			        document, textArea);
	             }  
	    });  
	    
	    //Translates the space character into an HTML non-breaking space
	    textArea.getInputMap().put (KeyStroke.getKeyStroke (' '), "nbsp" );  
	    textArea.getActionMap().put ("nbsp", new AbstractAction() {  
	      	     public void actionPerformed (ActionEvent e) {
	      	    	 textArea.replaceSelection ("\u00A0");      	    	 
	             }  
	    });
	    
	    //Set CTRL+Z shortcut for undo
	    textArea.getInputMap().put (KeyStroke.getKeyStroke ("control Z"), "undo2" );  
	    textArea.getActionMap().put ("undo2", new AbstractAction() {  
     	         public void actionPerformed (ActionEvent e) {  
     	    	     undo();    
     	         }  
        });
	    
	    //Set CTRL+C shortcut for customised copy
	    textArea.getInputMap().put (KeyStroke.getKeyStroke ("control C"), "copy2" );  
	    textArea.getActionMap().put ("copy2", new AbstractAction() {  
     	         public void actionPerformed (ActionEvent e) {  
     	    	     copy();    
            }  
        });
	    
	    //Set CTRL+V shortcut for customised paste
	    textArea.getInputMap().put (KeyStroke.getKeyStroke ("control V"), "paste2" );  
	    textArea.getActionMap().put ("paste2", new AbstractAction() {  
     	         public void actionPerformed (ActionEvent e) {  
     	    	     paste ();    
            }  
        });
	    
	    //Set CTRL+X shortcut for customised cut
	    textArea.getInputMap().put (KeyStroke.getKeyStroke ("control X"), "cut2" );  
	    textArea.getActionMap().put ("cut2", new AbstractAction() {  
     	         public void actionPerformed (ActionEvent e) {  
     	    	     copy ();
     	    	     replaceString ("", textArea.getSelectionStart(), textArea.getSelectionEnd(), 
		    	    		     document, textArea);    
            }  
        });
	    progress.setValue (18);
	    progress.paintImmediately (progress.getVisibleRect());
	    
	    //Sets up timer for auto-save. Carry out operations to strip accessibility features from text
	    saveTimer = new Timer (Global.AUTOSAVE_INTERVAL, new ActionListener () {
	        public void actionPerformed (ActionEvent e) {
	        	busy (true);
	        	saveDoc();
	        	busy (false);
	        } 
	    }); 
	    
	    //Set up on-demand refreshing thread
	    refresh = new Timer (0, new ActionListener () {
	        public void actionPerformed (ActionEvent e) {
	        	getWindow().getGlassPane().setCursor (Cursor.getPredefinedCursor (Cursor.DEFAULT_CURSOR));
	            busy(true);
	        	busy(false);
	        	getWindow().getGlassPane().setCursor (Cursor.getPredefinedCursor (Cursor.WAIT_CURSOR));
	        }
	    });
	    refresh.setRepeats (false);
	    		
	    //Initialises the undo list
		undo = new UndoManager();
				
		//Shutdown hook: saves any document left open and closes the database.
		Runtime.getRuntime().addShutdownHook (new Thread () {
			public void run () {
				if (isDocOpen())
				   closeDoc (true);	
				try {
					Global.DB_HANDLER.closeDataSource();
				} catch (SQLException eDB) {
					    System.out.println ("Error encountered when closing the database: " +eDB);
						eDB.printStackTrace();
				}				
			}
		});
		progress.setValue (19);
		progress.paintImmediately (progress.getVisibleRect());
	    
	    //Displays the login window
		splash.close();
		login.display();
				
		//Resets mouse pointer
	    busy (false);
	}
	
	/**
	 * Finds out which languages are available and presents them to the user through a 
	 * selection dialogue. If there exists a database file of a certain language (it has 
	 * already been set), the dialogue is not shown.
	 * @param owner Window which the login dialogue is displayed over
	 * @param splash Splash screen shown during editor startup
	 */
	public static void selectLanguage (Window owner, CanvasWindow splash) {
		final String[] dbFiles,			//database file list
		               localCodes;		//locale codes 
		final String sLanguage;			//selected language
		String sSysLanguage;			//system's default language
		final ImageIcon languageIcon;	//icon image for selection dialogue
		
		//Determines the number of existing database files 
	    dbFiles = new File (Global.INSTALL_FOLDER).list (new FilenameFilter() {                  
           				public boolean accept (File directorio, String fichero) {
           					return fichero.endsWith ("." +Global.DB_EXT);      
           				}
       				 }); 
	    
	    //If a previous database file exits, gets its language and use that one for the app
	    if (dbFiles.length == 1) {
	       Global.DB_FILE = dbFiles[0].substring (0, dbFiles[0].lastIndexOf ('.'));
	       localCodes = Global.DB_FILE.split ("_");
	       Global.DB_FILE = Global.DB_FILE +"."+ Global.DB_EXT;
	       Global.LANG = ResourceBundle.getBundle ("MessagesBundle", new Locale (localCodes[0], localCodes[1]), Thread.currentThread().getContextClassLoader());
	    
	    //No previous database file
	    } else if (dbFiles.length == 0) {
	    	   
	    	   //Determines the system's default language. If not recognised, use fallback language
		       sSysLanguage = Global.LANG.getLocale().getDisplayLanguage();
		       System.out.println ("Detected language: " +sSysLanguage);
		       if (!sSysLanguage.isEmpty()) {
		    	  sSysLanguage = Character.toUpperCase (sSysLanguage.charAt(0)) + sSysLanguage.substring(1);
		          if (!Global.LANGUAGES.containsKey (sSysLanguage))
		        	 sSysLanguage = Global.DEFAULT_LANG;
		       } else sSysLanguage = Global.DEFAULT_LANG;
		   
		       //Sets up the icon for the language selection dialogue
		       languageIcon = new ImageIcon (Global.IMG_FOLDER +Global.SEPARADOR_SYS+ Global.LANGUAGESET+ Global.PICT_EXT_PUNTO);
			   languageIcon.setImage (languageIcon.getImage().getScaledInstance (Global.ICON_SIZE, Global.ICON_SIZE, Image.SCALE_AREA_AVERAGING));
		       
			   //Sets up and displays the dialogue
			   if (splash != null)
			      splash.setVisible (false);
			   sLanguage = (String) JOptionPane.showInputDialog (owner, Global.LANG.getString ("idiomaMensaje"),
			                          Global.LANG.getString ("idiomaTitulo"), JOptionPane.PLAIN_MESSAGE,
			                          languageIcon, Global.LANGUAGES.keySet().toArray(), sSysLanguage);
			   if (splash != null) {
			      splash.setVisible (true);
			      splash.getCanvas().paintImmediately (splash.getCanvas().getVisibleRect());
			   }

		       //Updates the global language provided it has been confirmed. If not, it exits app.
			   if ((sLanguage != null) && (sLanguage.length() > 0))
				  Global.LANG = ResourceBundle.getBundle ("MessagesBundle", Global.LANGUAGES.get (sLanguage), Thread.currentThread().getContextClassLoader());   
			   else {
				    System.out.println (Global.LANG.getString ("idiomaError"));
				    System.exit (1);	
			   }
			   
			   //Copies the database file corresponding to the selected language
			   Global.DB_FILE = Global.LANG.getLocale().getLanguage() +"_"+ 
			                    Global.LANG.getLocale().getCountry()  +"."+ Global.DB_EXT; 
			   DataUsers.copyFile (new File (Global.DB_FOLDER +Global.SEPARADOR_SYS+ Global.DB_FILE), new File (Global.DB_FILE));
	
	    }
	    
	    //Updates the default locale for all new UI components. Geographical variants seem not to be
	    //supported (e.g: en_GB should be converted to en). Except for Portuguese, whose Brazilian variant
	    //is the only one supported.
	    if (Global.LANG.getLocale().getLanguage().equals ("pt"))
	       JComponent.setDefaultLocale (new Locale ("pt", "BR"));
	    else JComponent.setDefaultLocale (new Locale (Global.LANG.getLocale().getLanguage()));
	}
	
	/**
	 * Reads the current user's toolbar configuration file, creates all needed Swing buttons and
	 * groups them by type of action. An indexed list of those button is also set up.
	 * @return Object representing the buttons set as visible for the current user. 
	 */
	private CustomisedUI loadButtonConfig () throws Exception {
		CustomisedUI toolbars;
		JPanel menuPanel;
		int i;
		
		//Reads the toolbar config file
		toolbars = new CustomisedUI (sProfilePath+ Global.USER_CONF_XML, sProfilePath+ Global.USER_CONF_XSD);
		
		//Checks if pictograms are enabled. If not, resets pictograms button (in case it was on)
		bPictosEnabled = toolbars.getVertBar().isVisible ("pictogramas");
		if (!bPictosEnabled)
		   bPictosOn = false;
				
		//Builds the corresponding Swing buttons individually
		toolbars.getHorizBar().buttoniseVisibles (Global.ICON_SIZE, Global.ICON_SIZE,  
				                                 Global.IMG_FOLDER + Global.SEPARADOR_SYS, 
		                                         Global.PICT_EXT, "fuente", Global.FONTS_DISPLAYABLE, this);
		toolbars.getVertBar().buttoniseVisibles (Global.ICON_SIZE, Global.ICON_SIZE, 
				                                Global.IMG_FOLDER + Global.SEPARADOR_SYS,
                                                Global.PICT_EXT, "", null, this);
		
		//Merges the lists of visible buttons
		buttons = new HashMap<String, Button> ();
		buttons.putAll (toolbars.getHorizBar().getVisibles());
		buttons.putAll (toolbars.getVertBar().getVisibles());
		
		//Resets undo and copy/paste buttons
		if (buttons.containsKey("deshacer"))
		   buttons.get("deshacer").enable (false);
		if (buttons.containsKey("copiarOPegar"))
		   buttons.get("copiarOPegar").enable (false);
		
		//Sorts the buttons into groups and adds those to the appropriate toolbar
		toolbars.getHorizBar().groupVisibles (Global.G1_BUTTONS, Global.LANG.getString ("tituloGrupo1"), Global.BUTTON_EDGE);
		toolbars.getHorizBar().groupVisibles (Global.G2_BUTTONS, Global.LANG.getString ("tituloGrupo2"), Global.BUTTON_EDGE);
		toolbars.getHorizBar().groupVisibles (Global.G3_BUTTONS, Global.LANG.getString ("tituloGrupo3"), Global.BUTTON_EDGE);
		toolbars.getHorizBar().groupVisibles (Global.G4_BUTTONS, Global.LANG.getString ("tituloGrupo4"), Global.BUTTON_EDGE);
		toolbars.getVertBar().groupVisibles  (Global.G5_BUTTONS, Global.LANG.getString ("tituloGrupo5"), Global.BUTTON_EDGE);
		
		//If fonts enabled, replaces the font menu panel with the drop-down menu itself in the 
		//menu index.
		if (buttons.containsKey ("fuente")) {
		   menuPanel = (JPanel) buttons.get("fuente").getComponent();
	       for (i = 0; i < menuPanel.getComponents().length; i++)  
	           if (menuPanel.getComponents()[i] instanceof JComboBox) {
	              buttons.get("fuente").setComponent ((JComboBox) menuPanel.getComponents()[i]);
	           }	       
		}
		
		//Adds the credits and print buttons to the vertical toolbar.
		toolbars.getVertBar().getPanel().add (Box.createVerticalGlue());
		//toolbars.getVertBar().getPanel().add (printDoc);
		//toolbars.getVertBar().getPanel().add (Box.createRigidArea (new Dimension (0, Global.BUTTON_EDGE)));
		toolbars.getVertBar().getPanel().add (credits);
		
		return toolbars;
	}
	
	/**
	 * Returns the current user's ID.
	 * @return User ID
	 */
	public int getUsrId () {
		return iUsrId;
	}
	
	/**
	 * Returns the current user's name
	 * @return User name
	 */
	public String getUsrName () {
		return sUsrName;
	}
	
	/**
	 * Hides toolbars and the virtual keypad panel, resets buttons and disables the text area
	 */
	public void reset () {
		//Hides both the horizontal and vertical toolbars
		getTopPanel().setVisible (false);
		((BorderLayout) ((JPanel) getCentre()).getLayout()).getLayoutComponent (BorderLayout.WEST).setVisible (false);
		
		//Closes the PBDQ colour key dialogue and the virtual keypad panel if any was left open.
	    if (PBDQkey != null) {
		   PBDQkey.close();
	       getBottomPanel().setVisible (false);
	    }
	    
	    //Initialises all toggle buttons
	    bPictosOn = false;
	    bPBDQOn = false;
	    bVKeypadOn = false;
	    bCopy = false;        
		
		//Disables the text area.
		textArea.setBackground (Color.GRAY);
        textArea.setEnabled (false);        
	}
	
	/**
	 * Switches the current user and changes all user-dependent features of the window.
	 * @param iUsrId Switched-to user's id
	 * @param sUsrName Switched-to user's name
	 */
	public void loadUsr (int iUsrId, String sUsrName) {
		ImageArray avatarImage;	
		CustomisedUI toolbars;			
		JPanel centrePanel;			
				
		//Shows busy cursor
	    busy (true);
	    
	    //Is the user being switched to itself?
		if (this.iUsrId != iUsrId) {
		   		
		   //Is there any opened document?
		   if (isDocOpen())
			  closeDoc (false);
		   else if (this.iUsrId == 0)
			       
			       //Delete the welcome banner
				   try { 
					   textArea.getStyledDocument().remove (0, 1);
				   } catch (BadLocationException e1) { 
					       System.out.println ("Problema al suprimir la imagen de intro");
				   }	   
		   
		   //Resets the editor window
		   reset();
		   
		   //Updates all user-related data
		   this.iUsrId = iUsrId;
		   this.sUsrName = sUsrName;
		   sProfilePath = Global.USERS_FOLDER +Global.SEPARADOR_SYS+ iUsrId +Global.SEPARADOR_SYS;
		}
		
		//Updates avatar
		avatarImage = new ImageArray (Global.USER_AVATAR, sProfilePath, Global.PICT_EXT);
	    avatarImage.resize (Global.ICON_SIZE, Global.ICON_SIZE);
	    avatar.setIcon (avatarImage.toIcon());
	    avatar.setToolTipText (sUsrName);
		
		try {
			//Builds all toolbar buttons
			toolbars = loadButtonConfig();
			
			//Creates the directory where all RTF documents will be exported to
		    File dirDocs = new File (sProfilePath + Global.DOC_FOLDER);
		    if ((!dirDocs.exists()) && (!dirDocs.mkdir()))	
		       System.out.println ("Ha ocurrido un error al crear el directorio " +sProfilePath +Global.DOC_FOLDER);
		   
		    //Was the opening of the most recent document successful?
		    if (openMostRecent()) {
		       	 
		       //Updates the top horizontal toolbar
			   getTopPanel().remove(2);
		       getTopPanel().add (toolbars.getHorizBar().getPanel(), 2);
		       getTopPanel().revalidate();
		       getTopPanel().setVisible (true);
		          			  
		       //Updates the left-hand side vertical toolbar
		       centrePanel = (JPanel) getCentre(); 
		       centrePanel.remove (((BorderLayout) centrePanel.getLayout()).getLayoutComponent (BorderLayout.WEST));
			   centrePanel.add (toolbars.getVertBar().getPanel(), BorderLayout.WEST);
			   centrePanel.revalidate();
		    
		       //Updates the list of virtual keypads designed for this user
		       if (toolbars.getVertBar().isVisible ("tareas"))
		          loadKeypadMenu();
		       else getBottomPanel().setVisible(false);
		       
		       //There has been no change of user (no reset) => updates PBDQ and pictograms
		       if (this.iUsrId == iUsrId) {
		    	  //If pictogram feature has been disabled, switch it off.
			      if (!toolbars.getVertBar().isVisible ("pictogramas")) {
			         removePictos (document, textArea, false);
			         bPictosOn = false;
			      } else buttons.get("pictogramas").setSelected (bPictosOn); 
		    	   
		          //If PBDQ has been disabled, switch it off and close colour key
		          if (!toolbars.getVertBar().isVisible ("resalta")) {
		    	     pbdqParser (false, getBody(document).getStartOffset(), document.getLength() - 1, 
		    	    		     document, textArea);
		    	     bPBDQOn = false;
		    	     PBDQkey.close();
		          } else buttons.get("resalta").setSelected (bPBDQOn);	          
		       } 
		       
		       //A user has definately logged in => change login's title and exit mode
		       login.getWindow().setTitle (Global.LANG.getString ("cambioUsrTitulo"));
		       login.changeExitMode (JDialog.DISPOSE_ON_CLOSE);		       
		    } 
		    
		    //Resets the mouse pointer
		    busy (false);
		    
		//Problems when opening a config file or document? => display an error message and make login's "X" button definitive
		} catch (Exception e) {
			    System.out.println ("A problem has been encountered when opening the file: " +e);
                busy (false);
                if ((e.getMessage() == null) || !e.getMessage().equals ("newDoc cancel"))
			       JOptionPane.showMessageDialog (getWindow(), Global.LANG.getString ("loginErrorMensaje"), Global.LANG.getString ("loginErrorTitulo"), JOptionPane.ERROR_MESSAGE);
                else JOptionPane.showMessageDialog (getWindow(), Global.LANG.getString ("loginCancelMensaje"), Global.LANG.getString ("loginErrorTitulo"), JOptionPane.INFORMATION_MESSAGE);
                this.iUsrId = 0;
                login.changeExitMode (JDialog.EXIT_ON_CLOSE);
		        login.display();		        
	    }    
	}
		
	/**
	 * Check the similarity of two given colours according to a certain threshold. The latter is akin to 
	 * the maximum vector distance between two points in order to consider them as practically the same.
	 * @param color1 One of the colours to compare
	 * @param color2 The other colour
	 * @param threshold Maximum "distance" between colours
	 * @return True if they are similar. False if not.
	 */
	private boolean isColorSimilar (Color color1, Color color2, double threshold) {
		double colourDistance;
		
		colourDistance = Math.pow ((color1.getRed() - color2.getRed()), 2) +
                         Math.pow ((color1.getGreen() - color2.getGreen()), 2) +
                         Math.pow ((color1.getBlue() - color2.getBlue()), 2);
        if (colourDistance <= threshold) {
	       return true; 
        } else return false;
	}
	
	/**
	 * Updates the font colour of a certain region of text inside a given document. To that end, all nested FONT tags
	 * wholly contained in the region are deleted. It takes into account the two possible representations of
	 * colour: HTML.Tag.FONT->HTML.Attribute.COLOR and CSS.Attribute.COLOR.
	 * @param iSelStart Starting offset of the text selected for the colour change
	 * @param iSelEnd Ending offset of the text selected for the colour change
	 * @param newColor New font colour 
	 * @param document Representation of the text inside the editable area
	 * @param bPBDQ Is the colour changed because of a PBDQ-related action?
	 */
	private void setFontColour (int iSelStart, int iSelEnd, Color newColor, 
			                     HTMLDocument document, boolean bPBDQ) {
		int iPos;			        		//position of character(s)
		SimpleAttributeSet attrib, 			//text attributes
		                   attribFONT,	    //FONT tag attributes
		 				   attribBackup; 	//colour attributes backup
		Element elemento;
		
		attribBackup = null;
		
		//Is there a selection
		if (iSelEnd > iSelStart) {
		   
		   //Deletes any colour style included in the current selection
		   iPos = iSelStart;
		   while (iPos < iSelEnd) {
			   
			     //Colour can be expressed in two ways attribute-wise...
		         elemento = document.getCharacterElement (iPos);
		         attrib = new SimpleAttributeSet (elemento.getAttributes());
		         
		         //Backup of current attributes 
		         if (iSelEnd < elemento.getEndOffset())
		        	attribBackup = new SimpleAttributeSet (attrib);
		         
		         //Removes the HTML FONT->COLOR representation
		         if (attrib.isDefined (HTML.Tag.FONT)) {
		    	    attribFONT = (SimpleAttributeSet) attrib.getAttribute (HTML.Tag.FONT); 
		            attribFONT.removeAttribute (HTML.Attribute.COLOR);
		            if (attribFONT.getAttributeCount() == 0)
		        	   attrib.removeAttribute (HTML.Tag.FONT);
		            document.setCharacterAttributes (iPos, elemento.getEndOffset() - iPos, attrib, true);
		         }
		         
		         //Removes the CSS COLOR representation
		         if (attrib.isDefined (CSS.Attribute.COLOR)) {
		    	    attrib.removeAttribute (CSS.Attribute.COLOR);
			        document.setCharacterAttributes (iPos, elemento.getEndOffset() - iPos, attrib, true); 
		         }
		         
		         //Removes any attribute serving as a backup of the colour
		         //(only if the colour change is not PBDQ-related)
		         if (attrib.isDefined (HTML.Tag.SPAN) && !bPBDQ) {
		            attrib.removeAttribute (HTML.Tag.SPAN);
		            document.setCharacterAttributes (iPos, elemento.getEndOffset() - iPos, attrib, true); 
		         }
		            
		         //Recovers the attributes for any remaining characters not affected by the colour change
		         if (iSelEnd < elemento.getEndOffset())
		        	document.setCharacterAttributes (iSelEnd, elemento.getEndOffset() - iSelEnd, attribBackup, true); 
		         iPos =  elemento.getEndOffset();
		   }
		
		   //Changes the colour style for the selection
		   attrib = new SimpleAttributeSet ();
		   StyleConstants.setForeground (attrib, newColor);
		   document.setCharacterAttributes (iSelStart, iSelEnd - iSelStart, attrib, false);
		
		//No selection made => change colour from the current caret position onwards
		} else new StyledEditorKit.ForegroundAction (" ", newColor).actionPerformed(null);				
	}
		
	/**
	 * With the PBDQ feature on, it highlights all "p", "b", "d" or "q" characters present 
	 * in the text with different colours, each according to the current background colour. 
	 * A backup of the previous colour is made. This is used when the PBDQ feature is switched off. 
	 * @param bSwitchOn True si se desea activar la coloración de pbdq's. False para desactivar.
	 * @param document Representative model of the typed text. 
	 * @param textArea Editable area containing text.
	 */
	public void pbdqParser (boolean bSwitchOn, int iStartParsing, int iEndParsing, 
			                HTMLDocument document, JTextPane textArea) {
		int iIniPosCursor,			  	//caret's initial position		 
		    iCurrPosCursor;		  		//caret's current position	
		char character;				  	//character in text		
		Color newColour, 			  	//colour to be applied
		      currColour;			  	//current character colour
		SimpleAttributeSet attribSPAN, 	//attributes of <SPAN> tag
		                   attribFONT, 	//attributes of <FONT> tag
		                   attrib;	  	//attributes of text
		String sTITLEvalue;			  	//value for the TITLE attribute	
		
		iIniPosCursor = textArea.getCaretPosition();
		iCurrPosCursor = iStartParsing;
		textArea.setCaretPosition (iCurrPosCursor);		
		
		//For the whole document
		while ((iCurrPosCursor <= iEndParsing) && (iCurrPosCursor <= document.getLength())) {
		      try {
		          character = textArea.getText (iCurrPosCursor, 1).toLowerCase().charAt(0);
		          if ((character == 'p') || (character == 'b') || (character == 'd') || (character == 'q'))
		        	 
		        	 //is the current colour being changed?
		        	 if (bSwitchOn) {
		         		//is there a backup copy of the previous colour?
		         		attrib = new SimpleAttributeSet (document.getCharacterElement(iCurrPosCursor).getAttributes());
		         		if (!attrib.isDefined (HTML.Tag.SPAN)) { 
		         		   if (attrib.isDefined (CSS.Attribute.COLOR)) { //has any colour been defined at all?
		         			  //retrieve the current colour
		         			  currColour = StyleConstants.getForeground 
                                                              (document.getCharacterElement
                             	                                            (iCurrPosCursor).getAttributes());
		         			  sTITLEvalue = String.valueOf(currColour.getRGB());
		         		   } else sTITLEvalue = "defecto";    //no previous colour => default colour.
		         		   	         		   
		         		   //Saves the colour as a TITLE attribe for the SPAN tag
		         		   attrib = new SimpleAttributeSet();
		         		   attribSPAN = new SimpleAttributeSet();
		         		   attribSPAN.addAttribute (HTML.Attribute.TITLE, sTITLEvalue);
		              	   attrib.addAttribute (HTML.Tag.SPAN, attribSPAN);
		              	   document.setCharacterAttributes (iCurrPosCursor, 1, attrib, false);		          
		         		}
		         		
		              	//Changes the current colour
		            	setFontColour (iCurrPosCursor, iCurrPosCursor + 1, mapPBDQ.get(character)[iCurrBgColour], document, true); 
			         
		             //Back to the previous colour	
		        	 } else {  
			        	    attrib = new SimpleAttributeSet (document.getCharacterElement(iCurrPosCursor).getAttributes());
			        	    
			        	    if (attrib.isDefined (HTML.Tag.SPAN)) {
			        	       attribSPAN = new SimpleAttributeSet ((SimpleAttributeSet) attrib.getAttribute (HTML.Tag.SPAN));
			          	    
			        	       //The coloured characters had a previous color different from the default?
			        	       if (!attribSPAN.getAttribute(HTML.Attribute.TITLE).equals ("defecto")) { 
			          	    									   	
			          	          //Retrieves the previous colour 
			          	          newColour = new Color (Integer.valueOf ((String) attribSPAN.getAttribute (HTML.Attribute.TITLE)));
			          	       
			          	          //Previous colour similar to background colour? => change to default colour  
			          	          if (isColorSimilar (newColour, bgColours[iCurrBgColour], Global.COLOR_THRESHOLD))
			          	    	     newColour = fgColours.get (bgColours[iCurrBgColour]); 
			          	          setFontColour (iCurrPosCursor, iCurrPosCursor + 1, newColour, document, false);
			          	    
			          	       //Previous colour was default => removes info about colour => colour determined by stylesheet   
			        	       } else {
			          	    	      //Removes colour backup
			          	    	      attrib.removeAttribute (HTML.Tag.SPAN);				          	       
				          	       
				          	          //Removes any possible description of current colour
				          	          if (attrib.isDefined(CSS.Attribute.COLOR))
				          	             attrib.removeAttribute (CSS.Attribute.COLOR);
				          	          if (attrib.isDefined (HTML.Tag.FONT)) {
				 		    	         attribFONT = (SimpleAttributeSet) attrib.getAttribute (HTML.Tag.FONT); 
				 		                 attribFONT.removeAttribute (HTML.Attribute.COLOR);
				 		                 if (attribFONT.getAttributeCount() == 0)
				 		        	        attrib.removeAttribute (HTML.Tag.FONT);
				          	          } 
				          	       
				          	          //Applies changes
				          	          document.setCharacterAttributes (iCurrPosCursor, 1, attrib, true);			          	    	   			          	  
			          	       }
			        	    }
			         }
		          iCurrPosCursor++;
		          } catch (Exception e) {
		    	          System.out.println ("Error al cambiar color de pbdq: " +e);
		    	          e.printStackTrace();
		          }
		}
		textArea.setCaretPosition (iIniPosCursor);
	}
	
	/**
	 * Returns whether the current user has pictograms enabled in their profile or not.
	 * @return True if pictos active. False, if not.
	 */
	public boolean isPictosEnabled () {
		return bPictosEnabled;
	}
	
	/**
	 * Builds a label containing the pictogram representing a given word and a caption
	 * below that pictogram formatted in HTML. The label is made clickable if there
	 * is more than one pictogram possible for the word provided. 
	 * @param sHTMLWord HTML version of the word
	 * @param pictoIds Array of pictogram IDs that may represent the given word.
	 * @return Swing label with the icon and 
	 */
	private JLabel newPictoLabel (String sHTMLWord, final String[] pictoIds) {
		final JLabel pictoLabel;   //label containing icon representing the given word	
		ImageArray pictoImage;	   //image for the pictogram icon
		
		//Creates a label with the corresponding pictogram and the word below it
        pictoImage = new ImageArray (pictoIds[0], Global.PICT_FOLDER+ Global.SEPARADOR_SYS, Global.PICT_EXT);
		pictoImage.resize (iPictoSize, iPictoSize);
	  	pictoLabel = new JLabel (sHTMLWord, pictoImage.toIcon(), JLabel.CENTER);
	  	   
	  	//Sets its visual dimensions and other properties
	  	pictoLabel.setMaximumSize (new Dimension (iPictoSize + 10, iPictoSize));
	  	pictoLabel.setHorizontalTextPosition (JLabel.CENTER);
		pictoLabel.setVerticalTextPosition (JLabel.BOTTOM);
		pictoLabel.setBorder (BorderFactory.createEmptyBorder (2, 2, 2, 2)); 
		pictoLabel.setAlignmentY (iPictoSize / (float) (iPictoSize + 5));
		pictoLabel.setForeground (fgColours.get (bgColours[iCurrBgColour]));
		pictoLabel.setFont (new Font (Global.FONT_FAMILY, Font.PLAIN, Global.FONT_SIZE + 4));
		
		//Listen to mouse clicks only if there is more than one possible pictogram
		if (pictoIds.length > 1) {
		   
		   //Creates a tooltip that gives a hint of the number of possible pictograms 
		   pictoLabel.setToolTipText ((String.valueOf (pictoIds.length) +" "+ 
		              			      Global.LANG.getString("pictoTip")));
		   		   		
		   //Saves pictogram's state
		   pictoLabel.putClientProperty ("pictoIds", pictoIds);
		   pictoLabel.putClientProperty ("chosenIdx", 0);
		   pictoLabel.setCursor (Cursor.getPredefinedCursor (Cursor.HAND_CURSOR));
		   
		   //Change pictogram's state on mouse click.
		   pictoLabel.addMouseListener(new MouseAdapter(){
			   public void mouseClicked (MouseEvent e){
			       int pictoIndex = ((Integer) pictoLabel.getClientProperty("chosenIdx") + 1) % pictoIds.length;
				   pictoLabel.putClientProperty ("chosenIdx", pictoIndex);
				   ImageArray pictoImage = new ImageArray (pictoIds[pictoIndex], Global.PICT_FOLDER+ Global.SEPARADOR_SYS, Global.PICT_EXT);
				   pictoImage.resize (iPictoSize, iPictoSize);
				   pictoLabel.setIcon (pictoImage.toIcon());						
			   }
		   });
		   
		   //Make the pictogram's border visible on hover.
		   pictoLabel.addMouseListener (new MouseAdapter(){
			   public void mouseEntered (MouseEvent e) { 
				   pictoLabel.setBorder (BorderFactory.createLineBorder (Global.DEFAULT_BORDER, 2));
			   }
			   
			   public void mouseExited (MouseEvent e) { 
				   pictoLabel.setBorder (BorderFactory.createEmptyBorder (2, 2, 2, 2));
			   }
		   });
		}
		return pictoLabel;
	}
	
	/**
	 * Inserts the pictogram corresponding to a certain word, be it given or detected.
	 * If not given, detection is carried out one character at a time of the word at the
	 * caret's current position. Punctuation signs are left out. That is, they are not
	 * included in the pictogram's title.
	 * @param sWord Word to be represented by pictogram  
	 * @param iWordEnd Character offset where word ends.
	 * @return Number of characters replaced.
	 */
	public int insertPicto (String sWord, int iWordEnd) {
		String sHTMLWord,		   //HTML code the word is coded in
		       sEndQuestionMark;   //ending question/exclamation mark		       
		String[] pictoIds;		   //IDs of all pictograms matching the word 
		StringWriter out;		   //output buffer for string construction	
		int iWordEndPunc,		   //word's ending offset including punctuation  			
		    iWordStart,		       //start offset of the current word
		    iCharReplaced;		   //cumulative number of characters replaced 
		SimpleAttributeSet attrib; //current character's attributes
		
		textArea.removeCaretListener(this);
		iWordEndPunc = iWordEnd;
		sEndQuestionMark = "";		
		iCharReplaced = 0;
        try {
        	//Word detection
        	if (sWord == null) { 
        	   editorKit.runCommand("caret-previous-word");
               iWordStart = textArea.getCaretPosition();
               sWord = textArea.getText (iWordStart, iWordEnd - iWordStart);
                              
               //Caret is not at the end of word => search goes on
		  	   while (sWord.replaceAll("[\\p{Z}\\s]", "").matches ("[^" +Global.ALPHA_REGEX+ "0-9]")) { //contains no alphanumeric char?
		  		     
		  		     //Warns of an interrogative particle => remember to add a pictogram for it
		  		     if (sWord.matches("[!?]"))
		  		        sEndQuestionMark = sWord;
		  		     
		  		     iWordEnd = iWordStart;  					  
		  		     editorKit.runCommand ("caret-previous-word");
		  		     iWordStart = textArea.getCaretPosition();
		  		     sWord = textArea.getText (iWordStart, iWordEnd - iWordStart);
		  	   }
		  	
		  	//Word given => works out the start of the word and detects presence of !?   
        	} else {
        		
        		   //Is the last character of the word not an alphanumeric one? 
        		   //=> move end of word back one place
        		   if (textArea.getText(iWordEnd - 1, 1).matches("[^" +Global.ALPHA_REGEX+ "0-9]"))
        			  iWordEnd--;
        		   
        		   //Takes note of the question mark
        		   if (sWord.endsWith ("?") || sWord.endsWith ("!")) {
                      sEndQuestionMark = sWord.substring (sWord.length() - 1);
                      sWord = sWord.substring (0, sWord.length() - 1);
        		   }
        		   
        		   //Now it can determine where the word starts
        		   iWordStart = iWordEnd - sWord.length();
        	}
		  		  	                      	  	
        	//Inserts corresponding pictogram
		  	textArea.setCaretPosition (iWordEndPunc);
		  	pictoIds = Data.buscaItemsLista (sWord);
		  	if ((pictoIds != null) && (pictoIds.length > 0)) {
		  	   
		  	   //Backup the current attribute set once PBDQ attributes have been stripped out.
		       if (bPBDQOn)
		    	  preventInsPBDQ (textArea.getInputAttributes());
		  	   attrib = new SimpleAttributeSet (textArea.getInputAttributes().copyAttributes());
		  		
		  	   //Gets HTML string detected word is coded in (gets format information)
		  	   out = new StringWriter();
		  	   editorKit.write (out, document, iWordStart, iWordEnd - iWordStart);
	           sHTMLWord = out.toString().replaceAll ("\\s{2,}|\\n|\\r|</?body>|" +
	           		                                  "<p .*?>|</p>|</?li>|</?ol>|" +
	        		   								  "</?ul>|</?table>|</?tr>|</td>|" +
	        		   								  "<td .*?>", "").trim();    
		  	   
	           //Sets up and inserts the IMG tag representing the label. 
	           //Backs up the original HTML code being replaced by the pictogram in an ALT attribute
	           //The TITLE attribute contains the word without format (for quick reference and copying)
			   attrib.addAttribute (StyleConstants.NameAttribute, HTML.Tag.IMG);
			   attrib.addAttribute (HTML.Attribute.ID, Integer.toString(iAutoPictoID));
			   iAutoPictoID++;
			   attrib.addAttribute (HTML.Attribute.TITLE, sWord); 
			   attrib.addAttribute (HTML.Attribute.ALT, sHTMLWord.replaceAll("</?html>","").replace('"','¡'));
			   StyleConstants.setComponent (attrib, newPictoLabel (sHTMLWord, pictoIds));
			   			   
			   //Insert the pictogram after the question/exclamation mark (if applicable).
			   if (sEndQuestionMark.isEmpty())
				  document.insertString (iWordEnd, "\u00A0", attrib);
			   else document.insertString (iWordEnd + 1, "\u00A0", attrib);
			   		   
			   //If there is no question/exclamation mark, removes the word
		  	   if (sEndQuestionMark.isEmpty()) {
		  		  document.remove (iWordStart, iWordEnd - iWordStart);
		  		  iCharReplaced = iWordEnd - iWordStart;
		  	   }
		  	   
		  	//No pictogram available => move onto the punctuation mark 
		  	} else if (!sEndQuestionMark.isEmpty())
		  		      iWordStart = iWordEnd;
		  	   
		    //If there is a question/exclamation mark, add its pictogram
		  	if (!sEndQuestionMark.isEmpty()) {
		  	   pictoIds = ASyntax.TextRipper (sEndQuestionMark);
		  	 
		  	   //Is there indeed a pictogram for a question/exclamation mark on the db?
		  	   if ((pictoIds != null) && (pictoIds.length > 0)) {
		  		  
		  		  //Backup the current attribute set once PBDQ attributes have been stripped out.
			      if (bPBDQOn)
			    	 preventInsPBDQ (textArea.getInputAttributes());
		  		  attrib = new SimpleAttributeSet (textArea.getInputAttributes().copyAttributes()); 
		  		   
		  	      //Gets HTML string the mark is coded in (gets format information)	
		  	      out = new StringWriter();
	              editorKit.write (out, document, iWordEnd, 1);
	              sHTMLWord = out.toString().replaceAll ("\\s{2,}|\\n|\\r|</?body>|" +
	        		      							     "<p .*?>|</p>|</?li>|</?ol>|" +
	        		   								     "</?ul>|</?table>|</?tr>|</td>|" +
	        		   								     "<td .*?>", "").trim();
		  	   
	              //Sets up and inserts the the IMG tag representing the label			   
			      attrib.addAttribute (StyleConstants.NameAttribute, HTML.Tag.IMG);
			      attrib.addAttribute (HTML.Attribute.ID, Integer.toString(iAutoPictoID));
				  iAutoPictoID++;
			      attrib.addAttribute (HTML.Attribute.TITLE , sEndQuestionMark);
			      attrib.addAttribute (HTML.Attribute.ALT, sHTMLWord.replaceAll("</?html>","").replace('"','¡'));
			      StyleConstants.setComponent (attrib, newPictoLabel (sHTMLWord, pictoIds));
			   
			      //Bypass the pictogram of the word preceding the mark 
			      if (iWordEnd != iWordStart) 
				     document.insertString (iWordEnd + 2, "\u00A0", attrib);
			      
			      //? or ! after a word without pictogram => no pictogram to bypass
			      else document.insertString (iWordEnd + 1, "\u00A0", attrib);
				   
				  //Remove the question/exclamation mark (plus the word already represented)
				  document.remove (iWordStart, iWordEnd + 1 - iWordStart);
			      
			      iCharReplaced++;
		  	   }
		  	}
		  	
		    //Recovers text attibutes in force at positions before the inserted pictogram. 
		  	textArea.addCaretListener(this);
		  	preventInsPicto (textArea.getInputAttributes());		  			
		 				
		  	return iCharReplaced;
		} catch (Exception eTexto) {
			    textArea.addCaretListener(this);
		   	    textArea.setCaretPosition (iWordEndPunc);
		        System.out.println ("An error was encountered when inserting a pictogram: " +eTexto);
		        eTexto.printStackTrace();
			    return 0;
		}
	}
	
	/**
	 * Parses a piece of text starting at a given offset and inserts the pictograms 
	 * matching the words in the text. 
	 * @param iPictosStart Character offset which pictograms are inserted from
	 * @param iIniCaretPos Caret's initial position before inserting pictograms
	 * @return Number of characters previous to the caret's original position removed 
	 * from the document. 
	 */
	public int insertPictos (int iPictosStart, int iIniCaretPos) {
		int iCharDelta,			//number of characters added and/or removed 
		                        //previous to the caret's original position.
		    iCaretPos,			//caret's current position
		    iCharReplaced;	  	//number of characters replaced by last pictogram
		String sWord;			//word in the text
		char character;			//character inside a word
				
		iCaretPos = iPictosStart;
		iCharDelta = 0;
	    sWord = "";
	    	    
	    //Parse until reaching the end of text
	    busy (true);
	    while (iCaretPos <= document.getLength()) { 
	    	  try {
			   	  character = textArea.getText(iCaretPos, 1).charAt(0);
			
	    	      if ((!sWord.trim().replaceAll ("[\\p{Z}\\s]", "").isEmpty()) && 
	    		      (!Character.isDigit (character) && !Character.isLetter (character))) {
	    		  
	    		     //A punctuation mark has been found?
	    		     if ((character != ' ') && (character != '\u00A0')&& (character != '\n'))
                        iCaretPos++;  	    			 
	    	     
	    		     //A non-alphanumeric character has been found => finished detecting whole word
	    		     //Word and character are just an interrogation/exclamation mark? => sequence of marks => no pictogram
	    		     if (sWord.matches("[?|!]") && (sWord.charAt(0) == character))
	    			    iCharReplaced = 0;
	    		     else { 
	    			  
	    			      //Word followed by interrogation/exclamation mark? => add mark to word detected
	    			      if (Character.toString(character).matches("[?|!]"))
	    			         sWord = sWord + character;
	    			  
	    			      //Pictogram insertion: remove any beginning punctuation mark
	    			      iCharReplaced = insertPicto (sWord.replaceAll ("[\\p{Z}\\s\"'¿¡]", ""), iCaretPos);
	    	         }
	    		 
	    		     //If a picto has been inserted, update counter for removed characters
	    		     if (iCharReplaced != 0) {
	    				    			
	    			    //Caret's initial position inside the word replaced by picto? 
	    			    if ((iIniCaretPos > (iCaretPos - iCharReplaced)) && (iIniCaretPos < iCaretPos)) {
	    			       iCharDelta = iCharDelta + 1 + iCaretPos - iCharReplaced - iIniCaretPos;
	    			       iCaretPos = iCaretPos - iCharReplaced;
	    			       iIniCaretPos = iCaretPos;	    			    	
	    			   
	    			    //Caret's initial position beyond word replaced by picto?   
	    			    } else if (iIniCaretPos >= iCaretPos) {
	    					      iIniCaretPos = iIniCaretPos + 1 - iCharReplaced;
	    					      iCaretPos = iCaretPos - iCharReplaced;
	    					      if (iCaretPos <= iIniCaretPos) 
	    		    			     iCharDelta = iCharDelta + 1 - iCharReplaced;
	    					  
	    		        //Caret before word replaced by picto 
	    			    } else iCaretPos = iCaretPos - iCharReplaced;		    		    
	    		     }
	    		 
	    		     //Gets ready for next word 
	    	         sWord = "";	    	    
	    	      } else sWord = sWord + character;
	    	      iCaretPos++;
	    	  } catch (BadLocationException e) {
					  iCaretPos++;					  
			  }
	    }
	    busy (false);
	    return iCharDelta;	    
	}
	
	/**
	 * Removes all pictograms from the text by deleting those image tags with the "picto" ID.
	 * @param docModel Model representative of text
	 * @param textArea Editable text area associated with the document 
	 * @param bSaving True if removal of pictograms is triggered by the saving of document 
	 * being currrently edited.
	 */
	private void removePictos (HTMLDocument docModel, JTextPane textArea, boolean bSaving) {
		Iterator pictoIMG;		//element representing the IMG tag that holds the picto label
		int iPictoPos;			//picto's position
		String sHTMLWord; 		//HTML code for the word represented by picto.
		
		busy (true);
		
		//For every IMG with ID = picto...
		for (pictoIMG = document.getIterator (HTML.Tag.IMG); pictoIMG.isValid(); pictoIMG.next()) 
     	    if (pictoIMG.getAttributes().isDefined (StyleConstants.ComponentAttribute))
			   try {
				  //Retrieve the HTML the word was coded in.
				  sHTMLWord = ((JLabel)StyleConstants.getComponent (pictoIMG.getAttributes()))
				                                     .getText().replaceAll ("<html>|</html>", "");
				  
				  //Remove the pictogram.
				  iPictoPos = pictoIMG.getStartOffset();
				  docModel.remove (iPictoPos, 1);
				  
				  //Re-insert the word (including any format info) the pictogram was replacing.
			      if (iPictoPos == docModel.getParagraphElement(iPictoPos).getStartOffset()) 
		    	     docModel.insertAfterStart (docModel.getParagraphElement(iPictoPos), 
		    	    		                    sHTMLWord); 
		     	
		     	  //Inserts wherever the selection starts 
		          else editorKit.insertHTML (docModel, iPictoPos, "<SPAN></SPAN>" +sHTMLWord, 
		        		                     0, 0, HTML.Tag.SPAN);
			      
			      //If the pictogram contains a PBDQ character and leaving the PBDQ feature on
			      //were out of context, turn it off.
			      if ((bSaving || !bPBDQOn) && sHTMLWord.contains("span"))
			         pbdqParser (false, iPictoPos, 
			        		     iPictoPos + sHTMLWord.replaceAll("<.*?>","").length(), 
			        		     docModel, textArea);
			      
			      //Colours PBDQ characters of the pictogram's title if they aren't already
			      else if ((!sHTMLWord.contains("span")) && bPBDQOn && !bSaving) 
			    	      pbdqParser (true, iPictoPos, 
			        		          iPictoPos + sHTMLWord.replaceAll("<.*?>","").length(), 
			        		          docModel, textArea);
			      
			  } catch (Exception eLoc) {
				      System.out.println ("Error al eliminar pictogramas: " +eLoc);
				      eLoc.printStackTrace();
			  }		  
		
		busy (false);
	}   
	
	/**
	 * Backs up the attributes for every pictogram inside a given character offset range. The
	 * resulting list of attributes is indexed according to pictogram ID. 
	 * @param iStartOffset Start of the range being considered
	 * @param iEndOffset End of range being considered
	 * @return Indexed map of attributes 
	 */
	private HashMap<String, SimpleAttributeSet> backUpPictos (int iStartOffset, int iEndOffset) {
		HashMap<String, SimpleAttributeSet> pictoList;
        SimpleAttributeSet attrib;	//attribute set for a given element
	    Iterator IMGtags;			//for traversing all IMG tags present in affected paragraphs
		
		pictoList = new HashMap<String, SimpleAttributeSet>();
	    
		//For all IMG tags...   
        for (IMGtags = document.getIterator (HTML.Tag.IMG); IMGtags.isValid(); IMGtags.next()) {
     	    attrib = new SimpleAttributeSet (IMGtags.getAttributes().copyAttributes());
        	
     	    //... inside the given offset range and with a component attribute (i.e. pictos)
     	    if ((IMGtags.getStartOffset() >= iStartOffset) &&
     	        (IMGtags.getStartOffset() <= iEndOffset) &&
     	        attrib.isDefined (StyleConstants.ComponentAttribute)) 
     	       
     	       //Backs up all attributes defining the pictogram
     	       pictoList.put ((String) attrib.getAttribute(HTML.Attribute.ID), attrib);
        	
     	    //Discards the attributes.
     	    else attrib = null;
        }
		
		return pictoList;
	}
	
	/** 
	 * Fixes pictograms corrupted while changing the document's structure (especially lists).
	 * To that end, a list with the original attributes of all to-be-repaired pictograms must
	 * be provided. 
	 * @param pictoList List of pictograms' original attributes
	 * @throws BadLocationException
	 */
	private void fixPictos (HashMap<String,SimpleAttributeSet> pictoList) throws BadLocationException {
		int iStartPicto, iEndPicto;		//starting and ending offsets of pictogram
		
		//Find out pictogram's offsets, remove the corrupted pictogram and replace it. 
        for (String sID: pictoList.keySet()) {
        	iStartPicto = document.getElement(sID).getStartOffset();
        	iEndPicto = document.getElement(sID).getEndOffset();
        	document.remove (iStartPicto, iEndPicto - iStartPicto);
  	        document.insertString (iStartPicto, "\u00A0", pictoList.get (sID));	        		  
		}
	}
	
	/**
	 * Builds the menu of available virtual keypads by listing their respective configuration files.
	 * These are stored in the user profile's folder designated for virtual keypads.
	 */
	private void loadKeypadMenu () {
		String[] configFileList;		//list of virtual keypad configuration filenames  
		File vKeypad1Dir, vKeypad2Dir;	//objects representing the directories for virtual keypad files
		JMenuItem vKeypadItem;			//menu item representing one virtual keypad
		ImageArray iconTypeKeypad; 	    //icon besides each menu item indicating type of keypad
		int i;
		
		vKeypad1Dir = new File (sProfilePath+ Global.TCV1_FOLDER);
	    vKeypad2Dir = new File (sProfilePath+ Global.TCV2_FOLDER);
	    vKeypadMenu = new JPopupMenu ();
	    FilenameFilter xmls = new FilenameFilter() {
	        public boolean accept (File dir, String name) {
	            return name.endsWith(".xml");
	        }
	    };
     	   
	    try {
	        //Are there any configuration files of virtual keypads?
	        if ((vKeypad1Dir.list(xmls).length == 0) && (vKeypad2Dir.list(xmls).length == 0))
	    	   throw new Exception();
	    	
	    	//Retrieves the names of all config files for type-1 virtual keypads and adds them to the menu
	        configFileList = vKeypad1Dir.list (xmls);
	        iconTypeKeypad = new ImageArray ("tv1", Global.IMG_FOLDER + Global.SEPARADOR_SYS, Global.PICT_EXT);
	   		iconTypeKeypad.resize (60, 20);
	        for (i = 0; i < configFileList.length; i++) {  //configura la acción de cada item de menú 
	            vKeypadItem = new JMenuItem (configFileList[i].substring (0, configFileList[i].indexOf (".xml")),
	            		                       iconTypeKeypad.toIcon());
		        vKeypadItem.addActionListener (this);
		        vKeypadItem.setActionCommand ("tecladoVirtual1");
		        vKeypadMenu.add (vKeypadItem);
	        }    
	        
	        //Retrieves the names of all config files for type-2 virtual keypads and adds them to the menu
	        configFileList = vKeypad2Dir.list (xmls);
	        iconTypeKeypad = new ImageArray ("tv2", Global.IMG_FOLDER + Global.SEPARADOR_SYS, Global.PICT_EXT);
	   		iconTypeKeypad.resize (60, 20);
	        vKeypadMenu.addSeparator();
	        for (i = 0; i < configFileList.length; i++) { //Sets up the action for every menu item
	       	    vKeypadItem = new JMenuItem (configFileList[i].substring (0, configFileList[i].indexOf (".xml")),
	       	   		                        iconTypeKeypad.toIcon());
	            vKeypadItem.addActionListener (this);
	            vKeypadItem.setActionCommand ("tecladoVirtual2");
	            vKeypadMenu.add (vKeypadItem);
	        }	        
	    } catch (Exception eFichero) {  //Warns the user if no virtual keypads have been defined 
	    	    JOptionPane.showMessageDialog (super.getWindow(),
                             Global.LANG.getString ("tvMensajeTexto"), 
                             Global.LANG.getString ("tvMensajeTitulo"), 
                             JOptionPane.INFORMATION_MESSAGE);
	    	    buttons.get("tareas").enable (false);	            
	    }
	}
	
	/**
	 * Shows a panel at the bottom of the editor's window containing a given virtual keypad.
	 * The type of virtual keypad is specified through various parameters. 
	 * @param sVKeypadConfig Name of the configuration file for the virtual keypad to be displayed
	 * @param sCommand Action command string identifying the type of virtual keypad the source button belongs to.
	 * @param bVKeypad1 True if it is a type-1 virtual keypad. False if not.
	 */
	public void showKeypad (String sVKeypadConfig, String sCommand, boolean bVKeypad1) {
		JButton[] vKeypadButtons;			  
		JPanel vKeypadPanel;					
	    int i, iWindowWidth, iWindowHeight; 	//editor window's dimensions
        
	    bVKeypadOn = !bVKeypadOn;
        buttons.get("tareas").setSelected();
	      
	    //Finds out the window's dimensions
	    iWindowWidth = getCanvas().getHeight();
	    iWindowHeight = getCanvas().getWidth();
	    
	    //Retrieves the grid panel containing the virtual keypad.
	    if (bVKeypad1) {
	       TecladoVirtual1 teclado = new TecladoVirtual1 (sVKeypadConfig, iWindowHeight - 10, 
	    		                      (int) ((iWindowWidth - getTopPanel().getHeight()) * 0.5));
	       vKeypadButtons = teclado.getaBotonces();
	       vKeypadPanel = teclado.getCuadricula();
	    } else {
	    	   TecladoVirtual2 teclado = new TecladoVirtual2 (sVKeypadConfig, iWindowHeight - 10, 
	    			                     (int) ((iWindowWidth - getTopPanel().getHeight()) * 0.5),
	    			                     iUsrId);
	    	   vKeypadButtons = teclado.getaBotonces();
	    	   vKeypadPanel = teclado.getCuadricula();
	    }
                                            
	    //Sets the action command for every button in the keypad
        for (i = 0; i < vKeypadButtons.length; i++) {
		    vKeypadButtons[i].addActionListener (this);
		    vKeypadButtons[i].setActionCommand (sCommand);
	    }
	      
	    //Adds the keypad's grid panel to the editor's bottom area and displays it.
	    getBottomPanel().setVisible (false);
	    changeBottomPanel (vKeypadPanel);
	    getBottomPanel().setVisible (true);
	    getBottomPanel().revalidate();
	}
	
	/**
	 * Builds a list of all the user's documents and loads the most recent.
	 */
	private boolean openMostRecent () throws Exception {
		String[] fileList = null;
		String[] sRecentDoc;
		
		//No document has been loaded...
		if (!isDocOpen()) {
		
		   //Retrieves all user's documents. A file extension filter is used for that purpose.
		   fileList = new File (sProfilePath).list (new FilenameFilter() {                  
                       		public boolean accept(File dir, String fichero) {
                       			return fichero.endsWith ("." +Global.DOC_EXT);
                       		}
		           	  }); 
		
		   //Are there any document files? => determined the document last modified and opens it
		   sRecentDoc = DataUsers.getLastDocument (Integer.toString (iUsrId));
		   if ((fileList.length > 0) && (sRecentDoc != null)) { 
		      openDoc (sRecentDoc[0], sRecentDoc[1]);
		      return true;
		      
		   //No document file: the user has not created a document yet or ...
		   //... the database is new => ovewrites any previous file.
		   } else { 
		  	   
		   	      //Pictograms enabled? => Name document with a representative pictogram
			      if (bPictosEnabled) { 
		   	      		   		  
		   		     //Builds the dialogue with a pictogram selecction area and displays it.
				     NameDocPicto nombraDocPicto = new NameDocPicto (String.valueOf(iUsrId));
		   		     boolean aceptado = nombraDocPicto.muestraCampoConfirmar (getWindow());
				  				  				  
				     //Acts upon acceptance (loads the document) or cancellation
			         if (aceptado) {
			    	    newDoc (nombraDocPicto.getFields()[0], nombraDocPicto.getPictoId (Global.DEFAULT_PICT));					    				    
			         } else throw new Exception ("newDoc cancel");
			         return aceptado;
			      
			      //No pictogram on document naming    
		   	      } else { 
		   	    	  	 //Builds the dialogue and displays it.
				         NameDoc nombraDoc = new NameDoc (String.valueOf(iUsrId));
				         boolean aceptado = nombraDoc.muestraCampoConfirmar (getWindow());
				      				      
				         //Acts upon acceptance (loads the document) or cancellation
				         if (aceptado) {
				    	    newDoc (nombraDoc.getFields()[0], Global.DEFAULT_PICT);
						 } else throw new Exception ("newDoc cancel");
				         return aceptado;
			      }  
		      }
		
		//A document is already loaded
		} else return true;
	}
	
	/**
	 * Opens a given document file and loads it, resetting the auto-save timer. The window's title is
	 * changed, adding the name of the document. Additionally, the whole document's text is parsed in case
	 * any accessibility feature is enabled at the time of loading. Finally the undo facility is reset.
	 * @param sDocId Number that uniquely identifies the document and coincides with its filename without 
	 * extension. It effectively represents the document at the file system level.
	 * @param sDocName Name of the document. Represents it at the application level.
	 */
	public void openDoc (String sDocId, String sDocName) throws Exception {
		String sFilename;		//full name of the file to open
		Element body;			//root parent of the whole document 
				
		this.sDocId = sDocId;
		this.sDocName = sDocName;
		
		//Does the file exist?
		sFilename = sProfilePath +sDocId+ "." +Global.DOC_EXT;
		File docFich = new File (sFilename);
		if (docFich.exists()) {
		   		
		   //Retrieves the document model and the stylesheet associated with the text area
		   document = (HTMLDocument) editorKit.createDefaultDocument();
		   document.putProperty ("IgnoreCharsetDirective", Boolean.TRUE);
		   document.putProperty (HTMLDocument.StreamDescriptionProperty, null);
		   document.setPreservesUnknownTags (false);
		   document.addUndoableEditListener (this);
		   textArea.setDocument (document);
		   globalStyles = document.getStyleSheet();
				
		   //Loads the file into the document model
		   editorKit.read (new FileInputStream (docFich), document, 0);
		   
		   //If there is no empty paragraph inmediately after a table or a list, insert one
		   body = getBody (document);
		   if (!getNearestParent (document, body.getEndOffset(), "body table ul ol").getName().equals("body")) {
			  document.insertBeforeEnd (body, sEmptyParagraph);			  
  	       }
		   		   
		   //If applicable, adds inline features to the text 
		   if (bPictosOn) 
		      insertPictos (getBody(document).getStartOffset(), 0);
		   if (bPBDQOn)
		      pbdqParser (true, getBody(document).getStartOffset(), document.getLength() - 1, 
		    		      document, textArea);
		    
		   //Enables visual and interactive access to the text area  
		   textArea.setBackground (bgColours[iCurrBgColour]);
		   textArea.setCaretColor (fgColours.get(bgColours[iCurrBgColour]));
		   textArea.setEnabled (true);
		   textArea.grabFocus();
		   textArea.setCaretPosition (getBody(document).getStartOffset());
		   updateUI (textArea.getCaretPosition());
		   textArea.addCaretListener (this);
		   getWindow().setTitle (Global.LANG.getString("editorTitulo")+ " - " +sDocName);
		   		 		   
		   //Resets undo facility and auto-save timer
           undo.discardAllEdits(); 
           if (buttons.containsKey("deshacer"))
              buttons.get("deshacer").enable (false);
		   saveTimer.restart();
		      
		 //The given filename does not exist...
	     } else { 
		 	    JOptionPane.showMessageDialog (getWindow(),
			            sFilename +" "+ Global.LANG.getString ("errorFicheroTexto"), 
                        Global.LANG.getString ("errorFicheroTitulo"), 
                        JOptionPane.ERROR_MESSAGE);
				throw new IOException ("No existe el fichero");
		 }	
	}
	
	/**
	 * Dumps the contents of a given text area to a PNG image file with the same name as the document. This
	 * will be the previz image for its corresponding document.
	 * @param textToDump Text area whose contents are being dumped.
	 */
	private void savePreviz (JTextPane textToDump) {
		BufferedImage image;
		Graphics imageGraphics;
		
		//Graphic canvas for the image drawn from the text area's contents
	    image = new BufferedImage((int) (Global.SCREEN_DIM.height / 3.5), 
                                   (int) (Global.SCREEN_DIM.height / 2.625), 
                                   BufferedImage.TYPE_INT_RGB);
	    imageGraphics = image.getGraphics();
	    imageGraphics.setColor (Color.BLACK);
	    	    
	    //Create text area's graphical representation
	    //Set the dimensions of the hidden text area, making it displayable.
		//Necessary when generating the previz image (for painting its contents).
		textToDump.setSize (new Dimension (textArea.getWidth(), textArea.getHeight()));
	    textToDump.paint (imageGraphics);
	    	    
	    //Save graphical representation to the PNG file. 
	    try {
			ImageIO.write (image, "png", new File (sProfilePath +sDocId+ Global.PICT_EXT_PUNTO));
		} catch (IOException eIO) {
			    System.out.println ("Error al generar preview: " +eIO);
			    eIO.printStackTrace();
		}
	    imageGraphics.dispose ();
	}
	
	/**
	 * Launches the associated application for editing a given RTF document.
	 * @param sDocId Name at the storage level of the document whose RTF version the user
	 * whishes to edit.
	 * @param sDocName Name at the application level of the same document.
	 */
	public void launchRTF (String sDocId, String sDocName) {
		
		//Suspend autosave temporarily
		saveTimer.stop();
		
		try {
            File rtfDoc = new File (sProfilePath + Global.DOC_FOLDER + Global.SEPARADOR_SYS +
            		                sDocName +"."+ Global.DOC_EXPORT_EXT);
            File htmlDoc = new File (sProfilePath +sDocId+ "." +Global.DOC_EXT);
            
            //Is the document whose RTF version is being launched the same one being edited now?
            //=> update backup and re-generate RTF version
            if (sDocName.equals(this.sDocName)) 
               RTFConverter.ConvertFile (saveDoc(), rtfDoc.getAbsolutePath()); 
            
            //The HTML file exists => generate its RTF version now
            if (htmlDoc.isFile()) {
               String sHTML = DataUsers.fileToString (htmlDoc.getPath());
               sHTML = sHTML.replaceAll ("[\\n\\r](\\p{Z})*", "");
               RTFConverter.ConvertFile (sHTML, rtfDoc.getAbsolutePath());
            
            //There is no HTML file present => launching the RTF file is impossible
            } else JOptionPane.showMessageDialog (getWindow(), Global.LANG.getString ("nuevoRTFErrTexto"), 
            		                              Global.LANG.getString ("nuevoRTFErrTitulo"), 
            		                              JOptionPane.ERROR_MESSAGE);
            
            //If a windows system, launch associated app for .rtf from command line
            if (System.getProperty("os.name").toLowerCase().contains("windows")) {
      	       Runtime.getRuntime().exec ("rundll32 url.dll,FileProtocolHandler " + 
      	    		                       rtfDoc.getAbsolutePath());
      	       
      	    //It not a windows sys and Java Desktop supported, launch using edit method 
            } else Desktop.getDesktop().edit (rtfDoc);                    
  	    } catch (Exception e) {
  	          System.out.println ("Error when opening document: " +e); 
  	          e.printStackTrace();
  	    }
  	    
  	    //Resume autosave
  	    saveTimer.restart();
	}
		
	/**
	 * Creates a new document model, its respective file and adds the new document's name 
	 * to the window's title. It also resets the auto-save timer.
	 * @param sDocName New document's name. It will represent the document at the application level.
	 * @param sPictoFilename Image filename of the pictogram representing the new document.
	 */
	public void newDoc (String sDocName, String sPictoFilename) {
			
		//Shows busy cursor
		busy (true);
		
		//Sets up the document model
		sDocId = DataUsers.insertDocumentDB (sDocName, sPictoFilename, String.valueOf (iUsrId));
		this.sDocName = sDocName;
		document = (HTMLDocument) editorKit.createDefaultDocument();
		document.putProperty ("IgnoreCharsetDirective", Boolean.TRUE);
		document.setPreservesUnknownTags (false);
		document.addUndoableEditListener (this);
				
		//Sets default styles for new documents
		globalStyles = document.getStyleSheet();
		globalStyles.addRule ("body {color: #" +Integer.toHexString (Color.BLACK.getRGB() & 0x00ffffff)+ "}");
		globalStyles.addRule ("body {font-size: " +Global.FONT_SIZE+ "px; font-family: " +Global.FONT_FAMILY+ "}");
		globalStyles.addRule ("body {padding: 0px}");
		globalStyles.addRule ("table {" +Global.TABLE_BORDER+ "}");
		globalStyles.addRule ("td {" +Global.TD_BORDER+ "}");
				
		//Resets the text area and enables visual and interactive access to it
		textArea.setBackground (bgColours[iCurrBgColour]);
		textArea.setCaretColor (fgColours.get(bgColours[iCurrBgColour]));
		textArea.setEnabled (true);
		textArea.setDocument (document);
		
		//Guarantees the styles described in the current stylesheet are in effect 
		textArea.setText (textArea.getText());  
		
		//Update all interactive elements to reflect new document
		updateUI (textArea.getCaretPosition());
		textArea.grabFocus();
		textArea.addCaretListener (this);
		getWindow().setTitle (Global.LANG.getString("editorTitulo")+ " - " +sDocName);
				
		//Resets the undo facility and the auto-save timer
		undo.discardAllEdits(); 
		if (buttons.containsKey("deshacer"))
		   buttons.get("deshacer").enable (false);
	    saveTimer.restart();
	    
	    //Creates the document's file and previz.
	    try {
		    editorKit.write (new FileOutputStream (sProfilePath +sDocId+ "." +Global.DOC_EXT), document, 0, document.getLength());
		    textAreaCopy.setText ("");
		    savePreviz (textAreaCopy);	    
		} catch (Exception eIO) {
		        System.out.println ("Problemas al guardar el nuevo documento: " +eIO);
		  	    eIO.printStackTrace();
		}
	    	
	    //Resets mouse pointer
		busy (false);		
	}
	
	/**
	 * Makes a copy of the current document model and normalises such copy in the background 
	 * for subsequent rendering of the preview image and saving of the document file. This method 
	 * also ensures no WYSIWYG features appear in the preview (e.g. print margins).
	 * In the normalisation process, all pictograms are deleted, PBDQ changes reverted and 
	 * background-foreground colour scheme reset.  
	 */
	public String saveDoc () {
		String sHTML = "";			//HTML code corresponding to text area's contents
    	final int iPreviousColour;	//copy of the current background colour
    	HTMLDocument docModel;
    	
    	//Saves current background colour before changing it to default
		iPreviousColour = iCurrBgColour;
		iCurrBgColour = 0;
    	
    	try {
    		//Get rid of any spurious whitespace/line breaks.
    		//Replaces all IMG tags with the HTML backed up in ALT attribute
    		sHTML = textArea.getText().replaceAll ("[\\n\\r](\\p{Z})*", "");
    		sHTML = sHTML.replaceAll ("<img.*?alt=\"(.*?)\".*?>","$1");
    		sHTML = sHTML.replaceAll ("(<\\w.*?)¡(.*?)¡(.*?>)","$1\"$2\"$3");
    		textAreaCopy.setText (sHTML);
    		      				            		           	            
    		//Sets the stylesheet
    		docModel = (HTMLDocument) textAreaCopy.getDocument();
			docModel.getStyleSheet().addRule ("body {color: #" +Integer.toHexString (Color.BLACK.getRGB() & 0x00ffffff)+ "}");
    		docModel.getStyleSheet().addRule ("body {font-size: " +Global.FONT_SIZE+ "px; font-family: " +Global.FONT_FAMILY+ "}");
    		docModel.getStyleSheet().addRule ("body {padding: 0px}");
    		docModel.getStyleSheet().addRule ("li {color: #" +Integer.toHexString (Color.BLACK.getRGB() & 0x00ffffff)+ "}");
    		docModel.getStyleSheet().addRule ("table {" +Global.TABLE_BORDER+ "}");
    		docModel.getStyleSheet().addRule ("td {" +Global.TD_BORDER+ "}");
    		    			
    	    //Reverts PBDQ
    		if (bPBDQOn)
    		   pbdqParser (false, getBody(docModel).getStartOffset(), docModel.getLength() - 1,
    				       docModel, textAreaCopy);
    							    
    		//Sets the background back to the the default colour
    		textAreaCopy.setBackground (bgColours[iCurrBgColour]);
    							
    		//Updates the document's text file
    		editorKit.write (new FileOutputStream (sProfilePath +sDocId+ "." +Global.DOC_EXT), 
    		   		                               docModel, 0, docModel.getLength() + 1);
    		    
    		//Extracts the document's HTML contents.
    		sHTML =  textAreaCopy.getText().replaceAll ("[\\n\\r](\\p{Z})*", "");
    		    
    		//Guarantees the styles described in the current stylesheet are in effect 
    		textAreaCopy.setText (textAreaCopy.getText());
    		    
    		//Triggers the graphical dumping of the current contents to a PNG file = previz file. 
    		savePreviz (textAreaCopy);
    					
		} catch (Exception eGuarda) {
		        System.out.println ("Error encountered when normalising current document: " +eGuarda);
			    eGuarda.printStackTrace();		
		} finally {
			      iCurrBgColour = iPreviousColour;
		}
		   
		return sHTML;
	}
	
	/**
	 * Closes the document: saves the model representative of the current contents of the text area and 
	 * updates the database to mark the file as modified. The RTF version of the document is also generated.
	 * @param bExit Is the closing being performed because the application has exited?   
	 */
	public void closeDoc (boolean bExit) {
		final String sNormalHTML;	//HTML string obtained from normalised text
		
		//Shows busy cursor
		busy (true);
		
		saveTimer.stop();
		try {
			//Saves the document model
			document.removeUndoableEditListener (this);
			textArea.removeCaretListener (this);
			sNormalHTML = saveDoc();
							
			//Performs the conversion to RTF. A new thread is spawned if not exiting.
			if (!bExit) {
			   new Thread() { 
				    public void run () {
				    	RTFConverter.ConvertFile (sNormalHTML, sProfilePath + Global.DOC_FOLDER + 
	                            				  Global.SEPARADOR_SYS + sDocName +"."+ 
	                            				  Global.DOC_EXPORT_EXT); 
			        }		
			   }.start();
			} else RTFConverter.ConvertFile (sNormalHTML, sProfilePath + Global.DOC_FOLDER + 
  				                             Global.SEPARADOR_SYS + sDocName +"."+ 
				                             Global.DOC_EXPORT_EXT);
						
			//Updates the text area and disables it
			getWindow().setTitle (Global.LANG.getString ("editorTitulo"));	
			document.remove (0, document.getLength());
			textArea.setDocument (document);
			textArea.setEnabled (false);
			textArea.setBackground (Color.GRAY);
					    
			//Updates the document's modification date and blanks its ID.
			DataUsers.setUpdateDateDocument (sDocId);
			sDocId = "";	    
					
		} catch (Exception e) {
		        System.out.println ("Problemas al cerrar el fichero: " +e);
		        e.printStackTrace();
		}
		
		//Resets the mouse pointer
		busy (false);
	}
	
	/**
	 * Checks the document's ID. If empty, there is no open document.
	 * @return True if there is an open document. False if not.
	 */
	public boolean isDocOpen () {
		return !sDocId.isEmpty();
	}
	
	/**
	 * Returns the current document's name
	 * @return Document's name (does not represent it at the filesystem level)
	 */
	public String getDocName () {
		return sDocName;
	}
	
	/**
	 * Returns the document model's element corresponding to the BODY tag. The very first element of the 
	 * document model will not necessarily coincide with the first tag of the HTML code representing the
	 * document's contents. In that case, the remaining elements are displaced. Hence the need for this 
	 * method.   
	 * @param document Data model of the document 
	 * @return Element representing the BODY tag.
	 */
	private Element getBody (HTMLDocument document) {
		Element[] roots = document.getRootElements(); 
		Element body = null;
		for (int i = 0; i < roots[0].getElementCount(); i++) {
		    Element element = roots[0].getElement (i);
		    if (element.getAttributes().getAttribute (StyleConstants.NameAttribute) == HTML.Tag.BODY) {
		        body = element;
		        break;
		    }
		}
		return body;
	}
	
	/**
	 * Determines if the parent element of the current paragraph element is one of a number provided as a list
	 * of tag names. If so, it returns the parent element.
	 * @param document Document model of the text
	 * @param iCaretPos Caret's current position
	 * @param sTags Possible parent tags
	 * @return Parent element
	 */
	private Element getNearestParent (HTMLDocument document, int iCaretPos, String sTags) {
		Element raiz;
		
	    raiz = document.getParagraphElement (iCaretPos);
        do {
           raiz = raiz.getParentElement();
        } while (sTags.indexOf (raiz.getName()) ==  -1);
        return raiz;
	}
	
	/**
	 * Translates into HTML a given element of the document model.
	 * @param element Element to serialise to a HTML string
	 * @param out Serialiser to HTML string
	 * @return HTML string "equivalent" to given element
	 */
	private String extractHTML (Element element, StringWriter out) {
			
		ElementWriter writer = new ElementWriter (out, element);
        try {
			writer.write();
		} catch (IOException e) {
			    System.out.println ("Error encountered when serialising element: " +e);
			    e.printStackTrace();
		} catch (BadLocationException e) {
		        System.out.println ("Error encountered when extracting HTML at the element's position: " +e); 
			    e.printStackTrace();
		}
		return out.toString();
	}
	
	/**
	 * Retrieves the current font size from the given attibute set. 
	 * @param attrib Atribute set
	 * @return Font size
	 */
	protected int getFontSize (MutableAttributeSet attrib) {
		int tamano;
		
		tamano = StyleConstants.getFontSize (attrib);
		
		//If no font size attribute exists, getFontSize returns 12 by default => 18
		if ((tamano == 12) && !attrib.isDefined(StyleConstants.FontSize))
		   return 18; //TODO: why does it have to be 18 to make transition smooth?
		return tamano;		
	}
	
	/**
	 * Change the font size to the next bigger or smaller value according to the given map size. 
	 * Both the current and the new size is applied in pixels. Nevertheless, font size is ultimately
	 * represented in HTML form (1-7). 
	 * @param sizeMap Mapping of current size to would-be size.
	 */
	protected void setFontSize (HashMap<Integer, Integer> sizeMap) {
    	int i, 
    	    iDocStart,				//offset of the beginning of the document
    	    iParagraphEnd,			//ending offset of the paragraph being processed
    	    iSetIni, iSetEnd,		//starting and ending offset for applying new size
    	    iFirstChildSel,			//first character element included in selection
    	    iSelIni, iSelEnd,		//original starting and ending offset for applying new size
    	    iCurrentSize;			//current font size
    	MutableAttributeSet attrib; //text attributes
    	Element paragraph;	        //current paragraph
    	
    	iSelIni = textArea.getSelectionStart();
	    iSelEnd = textArea.getSelectionEnd();
	    textArea.setCaretPosition (iSelIni);
	    
	    //If selection exists, the set of affected character elements is traversed
	    if (iSelEnd > iSelIni) {
	       
	       //Initialise parameters for the traversing of elements
	       iDocStart = getBody(document).getStartOffset();
	       iParagraphEnd = iSelEnd;    
	       
	       //Traverse backwards, starting from the last paragraph included in the selection 
	       while ((iParagraphEnd >= iSelIni) && (iParagraphEnd >= iDocStart)) {
	    	  
	    	  //Find out the starting and ending character elements according to selection range 
	    	  paragraph = document.getParagraphElement (iParagraphEnd); 
	          i = paragraph.getElementIndex (iSelEnd);
	          iFirstChildSel = paragraph.getElementIndex (iSelIni);
	       
	          //With going backwards through every element...
	          while (i >= iFirstChildSel) {
	    	     
	    	        //Find out the current size
	    	        attrib = (MutableAttributeSet) paragraph.getElement(i).getAttributes();
	                iCurrentSize = getFontSize (attrib);
	             
	                //Set the new value for the size attribute according to current one 
	                attrib = new SimpleAttributeSet ();
	                StyleConstants.setFontSize (attrib, sizeMap.get (iCurrentSize));
	             
	                //Determines application range of new attribute value
	                iSetIni = paragraph.getElement(i).getStartOffset();
	                iSetEnd = paragraph.getElement(i).getEndOffset();
	                if (iSelIni > iSetIni)
	            	   iSetIni = iSelIni;
	                if (iSelEnd < iSetEnd)
	            	   iSetEnd = iSelEnd;
	            	
	                //Apply attribute value.
	                document.setCharacterAttributes (iSetIni, iSetEnd - iSetIni, attrib, false);
	                i--;
	          }
	          
	          //Determine the ending offset the the previous paragraph
	          iParagraphEnd = paragraph.getStartOffset() - 1;
	       }
	       
	    //If no selection, the font size attribute of the last typed character is changed accordingly
	    } else {
	    	   attrib = (MutableAttributeSet) textArea.getInputAttributes();
	    	   iCurrentSize = getFontSize (attrib);
	    	   new StyledEditorKit.FontSizeAction (" ", sizeMap.get (iCurrentSize)).actionPerformed(null);	    	  
	    }    
	}
	
	/**
	 * Inserts the given string, replacing any previously selected text. 
	 * The exception is when the selection includes several table cells, in which
	 * case only the text of the first cell is replaced 
	 * @param sString String to be inserted 
	 * @param iSelStart Start of selection
	 * @param iSelEnd End of selection.
	 * @param docModel Document model representation of the text
	 * @param textArea Editable text area associated with the document 
	 */
	protected void replaceString (String sString, int iSelStart, int iSelEnd,
			                      HTMLDocument docModel, JTextPane textArea) {
		Element startParagraph,		//paragraph element at selection's starting offset		
		        startSection, 		//parent element of previous
		        endParagraph,		//paragraph element st selection's ending offset
		        endSection;			//parent element of previous
		boolean bStartIsTD,			//selection start inside a table cell?
		        bEndIsTD,			//selection end inside a table cell?
		        bPartialTD;			//selection includes all table cells?
		int iStartFirstTable = 0,	//starting offset of first table included in selection 
		    iEndFirstTable = 0,		//ending offset of first table included in selection	
		    iStartLastTable = 0,	//starting offset of last table included in selection 
		    iEndLastTable = 0;		//ending offset of last table included in selection
		
	    //Caret's position is less than BODY tag's position? => Fix selection start value
	    if (iSelStart < getBody(docModel).getStartOffset())
	       iSelStart = getBody(docModel).getStartOffset();
	    
	    try {
	    	
	    	//If selection exists...
	    	if (iSelEnd - iSelStart != 0) {
	    	   startParagraph = docModel.getParagraphElement (iSelStart);
	    	   startSection = startParagraph.getParentElement();
	    	   endParagraph = docModel.getParagraphElement (iSelEnd);
	    	   endSection = endParagraph.getParentElement();
	    	   
	    	   //Selection start inside a table? => Find out table's beginning and end
	    	   bStartIsTD = startSection.getName().equals("td");
	    	   if (bStartIsTD) {
	    		  iStartFirstTable = startSection.getParentElement().getParentElement().getStartOffset();
	    		  iEndFirstTable = startSection.getParentElement().getParentElement().getEndOffset();
	    	   }	    		  
	    	   
	    	   //Selection end inside a table? => Find out table's beginning and end 
	    	   bEndIsTD = endSection.getName().equals("td");
	    	   if (bEndIsTD) {
	    		  iStartLastTable = endSection.getParentElement().getParentElement().getStartOffset();
	    		  iEndLastTable = endSection.getParentElement().getParentElement().getEndOffset();
	    	   }
	    	   
	    	   //Selection goes from one table to another and at least one of them is partially selected?
	    	   if (bStartIsTD && bEndIsTD)
	    	      bPartialTD = (iSelStart > iStartFirstTable) || (iSelEnd < (iEndLastTable - 1));
	    	   
	    	   //Selection starts in a table and includes it entirely?
	    	   else if (bStartIsTD && (iSelStart == iStartFirstTable) && (iSelEnd > (iEndFirstTable - 1)))
	    		       bPartialTD = false;
	    	   
	    	   //Selection ends in a table and includes it entirely?
	    	   else if (bEndIsTD && (iSelEnd == (iEndLastTable - 1)) && (iSelStart < iStartLastTable))
	    		       bPartialTD = false;
	    	   
	    	   //Unexpected => just in case, assume table has been partially selected
	    	   else bPartialTD = true;
	    	   
	    	   //Selection includes any table partially?
	    	   if ((iSelStart < endParagraph.getStartOffset()) && 
	    		   ((bStartIsTD && bEndIsTD && bPartialTD) || 
	    		    (bStartIsTD && bPartialTD && !bEndIsTD) || 
	    		    (bEndIsTD && bPartialTD && !bStartIsTD)))  
	    		  
	    		  //Change selection end so that just the the first paragraph inside the table is included. 
	    		  if (iSelEnd >= startParagraph.getEndOffset())
	    			 iSelEnd = startParagraph.getEndOffset() - 1;
	    		  else iSelEnd = startParagraph.getEndOffset();
	    	   
	    	   //Deletes selected piece of text
	    	   docModel.remove (iSelStart, iSelEnd - iSelStart);
	    	}
	    	
	    	//Prevent PBDQ colours spreading
	    	if (bPBDQOn)
	    	   preventInsPBDQ (textArea.getInputAttributes());	
	    	
	    	//Inserts the string
	    	docModel.insertString (iSelStart, sString, textArea.getInputAttributes());   	
	    		     	
		} catch (Exception eTexto) {
		        eTexto.printStackTrace();
		}
	}
	
	/**
	 * Inserts all HTML tags required to build an ordered/unordered list at the caret's current position. 
	 * If the aim is instead to turn the numbered/bulleted paragraphs into plain ones, it takes care of 
	 * deleting the necessary tags.
	 * @param sTypeList Type of list to build: "ul" or "ol". 
	 * @param iStartSel Starting offset of selected text.
	 * @param iEndSel Ending offset of selected text.
	 */
	protected void insertList (String sTypeList, int iStartSel, int iEndSel) {
		boolean bOnlyListSelected;			//selection includes a list exclusively					
		int iStartIndex, iEndIndex,  	    //element indexes included in selection 
		    iItemNo,						//total number of list items
		    i;
		String sHTML,        			    //HTML code of text represented by a given element
			   sHTMLBlock,					//HTML code block to be inserted into document model
			   sRest;						//part of the text remaining unselected after the selected block		        
		HTML.Tag tag;						//current list tag
		Element root, 						//root element of the document model tree
                section;					//element representing a block of text		        
	    SimpleAttributeSet attribIns;	    //backup of current input attributes	        
	    HashMap<String,SimpleAttributeSet> pictoList = new HashMap<String,SimpleAttributeSet>(); 
	    	    	    
	    //Notifies events about the present list building operation.
	    bBuildingList = true;
	    
	    //Finds the topmost parent element of the current paragraph (effectively, is the list inside a table?)
	    root = getNearestParent (document, textArea.getCaretPosition(), "td body");
	    
	    //Range of elements included in the selection
		iStartIndex = root.getElementIndex (iStartSel);
        iEndIndex = root.getElementIndex (iEndSel);
                
        //HTML-related initialisations
        sHTML = "";
        sHTMLBlock = "";
        tag = null;
                      
        //Checks if selection is comprised of just list items
        i = iStartIndex;
        bOnlyListSelected = true;
        do {
           tag = HTML.getTag (root.getElement(i).getName());
           
           //Is it a list tag?
           if ((tag == null) || ((!tag.equals (HTML.Tag.OL)) && (!tag.equals (HTML.Tag.UL))))
        	  bOnlyListSelected = false;
           i++;
        } while (bOnlyListSelected && (i <= iEndIndex)); 
        
        //Back up current input attributes
        attribIns = new SimpleAttributeSet (textArea.getInputAttributes());
      
        //Back up the attributes for every affected pictogram
        pictoList = backUpPictos (root.getElement(iStartIndex).getStartOffset(), 
        		                  root.getElement(iEndIndex).getEndOffset());
        
        try {
        	//At some point in the selection there is no previous list... 
            if (!bOnlyListSelected) {
                           	
               //Inserts <LI> tags for every text block
               for (i = iStartIndex; i <= iEndIndex; i++) {
            	   section = root.getElement(i);
            	   tag = HTML.getTag (section.getName());
            	              	   
            	   //Retrieves current HTML
            	   sHTML = extractHTML (section, new StringWriter());
                                      
                   //If it is non-listed text, reconstitute the paragraph
                   if (tag == null)
                	  sHTML = "<p style=\"margin-top: 0;\">" +sHTML+ "</p>";
                                 	  
                   //Text in a list already => no nesting (delete <UL>/<OL> tags)
                   if (sHTML.indexOf("<li>") != -1) { 
                	  sHTML = sHTML.substring (sHTML.indexOf("<li>"), sHTML.length());
                	  sHTML = sHTML.substring (0, sHTML.lastIndexOf("</li>") + 5);
                   
                   //Non-listed text => add <LI> tags	  
                   } else sHTML = "<li>" +sHTML+ "</li>"; 
                                      
                   sHTMLBlock = sHTMLBlock + sHTML;                  
               }
               sHTMLBlock = "<"+sTypeList+">" +sHTMLBlock.trim()+ "</"+sTypeList+">";
               
               //Gets the text coming after caret or end of selection
   			   sRest = textArea.getText (iEndSel, document.getLength() - iEndSel);
               
               //Adds an empty paragraph at the end of the list if the latter coincides with the end of the document
   			   //or if the rest of the document is empty. This is to avoid a glitch in the editor kit's write() method.
   			   //http://java-sl.com/tip_html_kit_last_empty_par.html               
               if ((root.getElement(iEndIndex).getEndOffset() == root.getEndOffset()) ||
                   sRest.replaceAll ("[\\p{Z}\\s]", "").trim().isEmpty())
                  sHTMLBlock = sHTMLBlock + sEmptyParagraph;
               
               //Removes the remaining old non-listed text block and saves resulting HTML string to document model
               document.setOuterHTML (root.getElement(iEndIndex), sHTMLBlock);
               if (iEndIndex > iStartIndex)
                  document.remove (root.getElement(iStartIndex).getStartOffset(), 
   		                           root.getElement(iEndIndex - 1).getEndOffset() - 
   		                           root.getElement(iStartIndex).getStartOffset());
                                          
            //Selection just includes list items
            } else {
            	              	   
            	   //Works out the list's length in terms of element indexes
            	   root = root.getElement (root.getElementIndex (iStartSel));
            	   iItemNo = root.getElementCount();
            	   iStartIndex = root.getElementIndex (textArea.getSelectionStart());
                   iEndIndex = root.getElementIndex (textArea.getSelectionEnd());
                   
                   //For everery <LI> block, remove the <LI> tag
                   for (i = iStartIndex; i <= iEndIndex; i++) {
                       sHTML = extractHTML (root.getElement(i), new StringWriter());        
                       sHTML = sHTML.substring(sHTML.indexOf("<li>") + 4, sHTML.length());
                       sHTML = sHTML.substring(0, sHTML.lastIndexOf("</li>"));
                       sHTMLBlock = sHTMLBlock + sHTML;                      
      	           }
                   
                   //List selected partially? => divide list
                   if (iItemNo > (iEndIndex - iStartIndex + 1)) {
                	                  	                                            
                      //Saves HTML string to document model
                      editorKit.insertHTML (document, root.getElement(iEndIndex).getEndOffset(), 
                    		                sHTMLBlock, 3, 0, HTML.Tag.P);
                                               
                      //Removes the old block 
                      document.remove (root.getElement(iStartIndex).getStartOffset(), 
                	     	           root.getElement(iEndIndex).getEndOffset() - 
                	     	           root.getElement(iStartIndex).getStartOffset());
   	                                     
                   //Elimina el tag de lista correspondiente asociado a todo el bloque	   
                   } else document.setOuterHTML (root, sHTMLBlock.trim());      	           
            }
            
            //Fixes pictograms corrupted during the insertion/deletion of lists
            fixPictos (pictoList);
            
        } catch (Exception eTexto) {
	            System.out.println ("Problemas al eliminar/insertar texto: " +eTexto);
	            eTexto.printStackTrace();
        }
               
        //Recover selection. Previous operations displace the cursor and thus selection highlight is lost
        textArea.setSelectionStart (iStartSel);
        textArea.setSelectionEnd (iEndSel);

        //If only one list item has been created and is the first one, copy all previous style information to the list
        if ((!bOnlyListSelected) && (iStartSel == iEndSel)) {
       	   textArea.setCharacterAttributes (attribIns, false); 
        }
                
        //Notifies the end of list building to events.
        bBuildingList = false;       
	}
	
	/**
	 * Generates all necessary tags for inserting a table with a given number of rows and columns
	 * (the latter with a specified width) at the caret's current position.
	 * @param iRows Number of rows for the new table
	 * @param iCols Number of columns for the new table
	 * @param iWidth Column width for the new table
	 */
	protected void insertTable (int iRows, int iCols, int iWidth) {
		Element currentParagraph,	//paragraph at the caret's current position
                tableParent;		//immediate parent element of table
	    String sHTMLTable, 			//HTML string representing the table
	           sRest,				//remaining text from the current position
	           sParContents;		//text contents of the current paragraph
	    boolean bEmptyParagraph;	//is the current paragraph empty?
	    int iCaretPos,				//caret's current position
	    	i, j;				
	    
	    iCaretPos = textArea.getCaretPosition();
	    
	    //Generates the HTML code required for the table except the closing tag (</TABLE>)
		sHTMLTable = "<table>";
        for (i = 0; i < iRows; i++) {
            sHTMLTable = sHTMLTable + "<tr>";
            for (j = 0; j < iCols; j++) 
   	            sHTMLTable = sHTMLTable + "<td width=\"" +iWidth+ "\">" +sEmptyParagraph+ "</td>"; 
            sHTMLTable = sHTMLTable + "</tr>";
        }

        try {
       	    //Determines if the paragraph that includes the caret's current position is empty
       	    currentParagraph = document.getParagraphElement (iCaretPos);
			sParContents = textArea.getText (currentParagraph.getStartOffset(), 
			                               currentParagraph.getEndOffset() - currentParagraph.getStartOffset());
			bEmptyParagraph = sParContents.replaceAll ("[\\p{Z}\\s]", "").trim().isEmpty();
		 
			//Gets the text that remains from the caret's current position onwarad
			sRest = textArea.getText (iCaretPos, document.getLength() - iCaretPos);
						
          	//No text after the table? => insert ending empty paragraph.            
            tableParent = getNearestParent (document, iCaretPos, "td li body");
            if ((currentParagraph.getEndOffset() == tableParent.getEndOffset()) ||
                (sRest.replaceAll ("[\\p{Z}\\s]", "").trim().isEmpty()))
               sHTMLTable = sHTMLTable + "</table>" +sEmptyParagraph;
            else sHTMLTable = sHTMLTable + "</table>";
 
            //If the paragraph where the table is to be inserted is empty, do the insertion replacing it
            if (bEmptyParagraph)
   	           document.setOuterHTML (currentParagraph, sHTMLTable);
            
            //If not, insert table after the paragraph
       	    else if (currentParagraph.getEndOffset() != tableParent.getEndOffset()) 
       	    	    document.insertAfterEnd (currentParagraph, sHTMLTable);
       	         else editorKit.read(new StringReader (sHTMLTable), document, currentParagraph.getEndOffset() - 1);
            
            //Sets the caret's position to its original 
            textArea.setCaretPosition (iCaretPos);
        } catch (Exception eTexto) {
                System.out.println ("Problemas al eliminar/insertar tag de tabla: " +eTexto);
                eTexto.printStackTrace();
        }                   
	}
	
	/**
	 * Inserts at a given position an IMG tag referencing a given image file. The image 
	 * replaces any previously selected text if applicable. 
	 * @param sImagePath Path to the the image file
	 * @param iSelStart Starting offset of the current selection
	 * @param iSelEnd Ending offset of the current selection.
	 */
	protected void insertImg (String sImagePath, int iSelStart, int iSelEnd) {
		SimpleAttributeSet attrib;	//attribute set at insertion position
		String sHTML;				//HTML code referencing the image file.
		
		try {
			//Backs up current input attributes before insert	       	    	
	    	attrib = new SimpleAttributeSet (textArea.getInputAttributes());
   		 
	       	//Removes any previously selected text
	       	replaceString ("", iSelStart, iSelEnd, document, textArea); 
	       	
	       	//Builds the HTML code referencing the image. 
   		    sHTML = "<IMG SRC=\"file:" + sImagePath + "\" hspace=\"2\" vspace=\"2\"/>";
   		 
   		    //Selection starts just where a paragraph does? => insert exactly at the beggining of paragraph	
   		    if (iSelStart == document.getParagraphElement(iSelStart).getStartOffset()) 
   	           document.insertAfterStart (document.getParagraphElement(iSelStart), sHTML); 
    	
    	    //Inserts wherever the selection starts 
            else editorKit.insertHTML (document, iSelStart, sHTML, 0, 0, HTML.Tag.IMG);
   		 
   		    //Recovers input attributes before inserting the image
   		    textArea.setCharacterAttributes (attrib, false);
   		 
        } catch (Exception eTexto) {
		         System.out.println ("Problemas al eliminar/insertar tag de imagen: " +eTexto);
		         eTexto.printStackTrace();
	     }
	}
	
	/**
	 * Carries out the undo action. If there are no more actions that can be undone, the undo button is disabled.
	 */
	public void undo() {
		try {
	    	if (undo.canUndo())
	           undo.undo();
	    	else buttons.get("deshacer").enable (false);
	    } catch (CannotUndoException eUndo) {  //se ha deshecho todo?
	            buttons.get("deshacer").enable (false);
	    	    JOptionPane.showMessageDialog (super.getWindow(),
	    	                                   Global.LANG.getString("deshacerTexto"), 
				                               Global.LANG.getString ("deshacerTitulo"), 
				                               JOptionPane.INFORMATION_MESSAGE);
	    }
	}
	
	/**
	 * Performs the copy action and updates the appearance of the copy/paste button accordingly
	 */
	public void copy () {
		Clipboard clip;			//system's clipboard 
		StringWriter out;		//output buffer for string construction	
		String sTextToCopy;		//string to be copied to clipboard
				
		out = new StringWriter();
		try {
        	//Extract html code corresponding to selected text
			editorKit.write (out, document, textArea.getSelectionStart(), 
					         textArea.getSelectionEnd() - textArea.getSelectionStart());
			sTextToCopy = out.toString();
			
			//Remove all tags except the <IMG> ones representing pictograms
			sTextToCopy = sTextToCopy.replaceAll ("(?!<img.*?id=.*?>)<.*?>", "");
			
			//Replace the <IMG> tag with the value of its "title" attribute
			sTextToCopy = sTextToCopy.replaceAll ("<.*?title=\"(\\w*|[!?])\".*?>","$1");
						
			//Replace nbsp entities and line breaks with spaces
			sTextToCopy = sTextToCopy.replaceAll ("([\\n\\r])(\\p{Z})*", "$1");
			sTextToCopy = sTextToCopy.replaceAll ("\r", "");
			sTextToCopy = sTextToCopy.replaceAll ("&#160;|\\n", " ").trim();
			
			//Replace other entities with their corresponding character
			sTextToCopy = RTFConverter.cleanHTMLCode (sTextToCopy);
						
			//Transfer to clipboard
			clip = Toolkit.getDefaultToolkit().getSystemClipboard();
	   	    clip.setContents (new StringSelection (sTextToCopy), null);

		} catch (Exception e) {
			    System.out.println ("Clipboard error: " +e);
			    e.printStackTrace();
		} 	
		
		((JButton) buttons.get("copiarOPegar").getComponent()).setIcon (pasteIcon);
	    bCopy = true;    	
	}
	
	/**
	 * Adapt any previously selected and copied text before performing the corresponding paste.
	 * It also updates the appearance of the copy/paste button.
	 * If the copied passage cannot be converted to plain text,the copying is not performed. 
	 */
	public void paste () {
   	    Clipboard clip;			//system's clipboard 
   	    Transferable copied;	//contents of the clipboard
   	    String sTextToPaste;	//string to be pasted from clipboard
   	    int iSelStart,          //starting offset of selection
   	        iCaretPosAfter;		//caret's position after inserting text
   	    
   	        
   	     	    
   	    //Find out the offset of current selection's start.
   	    iSelStart= textArea.getSelectionStart();
   	    
   	    //Check that some text has indeed been copied
   	    clip = Toolkit.getDefaultToolkit().getSystemClipboard();
   	    copied = clip.getContents (this);
   	    if (copied != null) {
   	       try {
   	    	   
   	    	   //Regardless of its source format, copied text is reduced to its plain text version
   	    	   if (copied.isDataFlavorSupported (DataFlavor.stringFlavor)) { 
   		    	 
			      //Remove breaks, convert whitespaces and paste.
   		       	  sTextToPaste = (String) clip.getData (DataFlavor.stringFlavor);
   		    	  sTextToPaste = sTextToPaste.replaceAll ("\\n|\\r|\\s", "\u00A0");
   		    	      
   		    	  //Insert copied text replacing current selection.
   		    	  replaceString (sTextToPaste, iSelStart, 
   		    			         textArea.getSelectionEnd(), 
   		    	   		         document, textArea);
   		    	  
   		    	  //Parse the newly inserted text for PBDQ (if on)
   		    	  if (bPBDQOn) 
   		    		 pbdqParser (true, iSelStart, iSelStart + sTextToPaste.length(), 
   	        		             document, textArea);
   		    	  
   		    	  //Insert pictograms for the newly added text if the picto feature is on
   		    	  if (bPictosOn) {
   	        	     iCaretPosAfter = textArea.getSelectionEnd(); 
   	        	     textArea.setCaretPosition (insertPictos (iSelStart, iCaretPosAfter) + iCaretPosAfter);
   	              }   		    	  
   		    	
   		       //If the copied text cannot be turned into a string, throw an exception.
   		       } else throw new Exception();
		   } catch (Exception eClipBoard) {
			       System.out.println ("Error al extraer del portapapeles: " +eClipBoard);
			       eClipBoard.printStackTrace();
		   }
		   
		   //Changes the button's icon accordingly and disables it. 
		   ((JButton) buttons.get("copiarOPegar").getComponent()).setIcon (copyIcon);
	       bCopy = false;
	       buttons.get("copiarOPegar").enable (false);
   	    }
	}
	
	/**
	 * Prevents inserting an IMG tag representing a picto when this is not really intended
     * It occurs when a new character is being inserted near a pictogram: the attributes
     * that define the tag for the pictogram may be "transfered" to other positions by
     * means of the input attribute facility.	
	 * @param attribIns Input attributtes at caret's current position
	 * @return The attribute set in question with pictogram attributes removed. Returns null 
	 * if the attribute set does not represent a pictogram.
	 */
	private MutableAttributeSet preventInsPicto (MutableAttributeSet attribIns) {
		if (attribIns.isDefined (StyleConstants.ComponentAttribute)) {
  		   attribIns.removeAttribute (StyleConstants.ComponentAttribute);
  	       attribIns.addAttribute (StyleConstants.NameAttribute, HTML.Tag.CONTENT);
  	       attribIns.removeAttribute ("$ename");
  	       attribIns.removeAttribute (HTML.Attribute.TITLE);
  	       attribIns.removeAttribute (HTML.Attribute.ID);
  	       attribIns.removeAttribute (HTML.Attribute.ALT);
  	       return attribIns;
        } else return null;
	}
	
	/**
	 * Prevents the attribute set corresponding to a coloured PBDQ character being carried over 
	 * to the next inserted character. This correction is needed when adding new text inmediately
	 * after a 'p', 'b', 'd' or 'q' and the PBDQ mode is on. 
	 * @param attribIns Input attributtes at caret's current position
	 */
	private void preventInsPBDQ (MutableAttributeSet attribIns) {
		MutableAttributeSet attribSPAN;  //attributes of <SPAN> tag
		
		if (attribIns.isDefined (HTML.Tag.SPAN) && bPBDQOn) {
       	   attribIns.removeAttribute (CSS.Attribute.COLOR);
       	   attribSPAN = (SimpleAttributeSet) attribIns.getAttribute (HTML.Tag.SPAN);
           if (!((String) attribSPAN.getAttribute(HTML.Attribute.TITLE)).equals("defecto"))
               StyleConstants.setForeground (attribIns, new Color (Integer.valueOf ((String) 
           		                                              attribSPAN.getAttribute 
           		                                              (HTML.Attribute.TITLE))));
       	   attribIns.removeAttribute (HTML.Tag.SPAN);		             		             
        }		     
	}
	
	/**
	 * Manages events coming from the editor's buttons.
	 */
	@SuppressWarnings("serial")
	public void actionPerformed (ActionEvent e) {
		final int iCaretPos,		//caret's original position
		          iSelStart,		//starting offset of selection
		          iSelEnd;  		//ending offset of selection
		
		iSelStart = textArea.getSelectionStart();
		iSelEnd = textArea.getSelectionEnd();
		iCaretPos = textArea.getCaretPosition();
						
		//New/Open document dialogue. TODO: decouple document update from instantiation.
		if (e.getSource() == newDoc) {
		   SelectDoc eligeNuevo = new SelectDoc (Global.LANG.getString ("nuevoTitulo"),
				                                                 20, 80, this, login);
		   eligeNuevo.display(getWindow());
					      
	    //Font type drop down menu. 
	    } else if (e.getActionCommand().equals ("fuente")) {
	    	      String sNombreFuente = ((JComboBox) e.getSource()).getSelectedItem().toString();
                  SimpleAttributeSet attrib = new SimpleAttributeSet();
	              StyleConstants.setFontFamily (attrib, sNombreFuente);
	              textArea.setCharacterAttributes (attrib, false);
	              
	    //Font colour selection.
	    } else if (e.getActionCommand().equals ("color")) {
	    	      Color newColour = null;
	    	      	    	      
	    	      //There is a selection => no colour 
	    	      if (iSelEnd > iSelStart)
	    	    	  colorPane.setColor (null); 
	    	      
	    	      //No selection => get colour from input attributes
	    	      else colorPane.setColor (StyleConstants.getForeground (textArea.getInputAttributes()));
	    	      
	    	      //Displays colour selection dialogue and waits for user's choice
	    	      colorChooser.setVisible (true);
	    	      newColour = colorPane.getColor();
	    	     	    	      
	    	      //Applies the selected colour if it's different from background
	    	      if ((newColour != null) && !isColorSimilar (newColour, bgColours[iCurrBgColour], 
	    	    		                                      Global.COLOR_THRESHOLD))
	    	    	 setFontColour (iSelStart, iSelEnd, newColour, document, false);	    	    	 
	    	   	    	      
	    //Increase font size button
	    } else if (e.getActionCommand().equals ("aumentar")) { 
	    	      setFontSize (new HashMap<Integer, Integer>() {{
	    	            put (8, 10);  put (10, 12); put (12, 14);
	    	            put (14, 18); put (18, 24); put (24, 36);
	    	            put (36, 36);
	    	        }});
	    	      textArea.setCaretPosition (iCaretPos);
	    	      textArea.setSelectionStart (iSelStart);
	    	      textArea.setSelectionEnd (iSelEnd);
	    	      
	    //Decrease font size button
	    } else if (e.getActionCommand().equals ("disminuir")) {
	    	      setFontSize (new HashMap<Integer, Integer>() {{
	    	    	    put (36, 24); put (24, 18); put (18, 14);
	    	    	    put (14, 12); put (12, 10); put (10, 8);
	    	    	    put (8, 8);
	    	    	}});
	    	      textArea.setCaretPosition (iCaretPos);
	    	      textArea.setSelectionStart (iSelStart);
	    	      textArea.setSelectionEnd (iSelEnd);
	    
	    //Bold button. Uses editor kit's pre-made action. 
	    } else if (e.getActionCommand().equals ("negrita")) {
	    	      editorKit.runCommand ("font-bold");
	    	      buttons.get("negrita").setSelected();
	    	      
	    //Italic button. Uses editor kit's pre-made action.
	    } else if (e.getActionCommand().equals ("cursiva")) {
	    		  editorKit.runCommand ("font-italic");
	    	      buttons.get("cursiva").setSelected();
	    	      
	    //Underline button. Uses editor kit's pre-made action.
	    } else if (e.getActionCommand().equals ("subrayado")) {
	    		  editorKit.runCommand ("font-underline");
	    	      buttons.get("subrayado").setSelected();
	    	      
	    //Insert image button.
	    } else if (e.getActionCommand().equals ("insertarImagen")) {
		    	  //Switch to system L&F. Couldn't do it before due to localisation issues with
	    	      //names of system folders.
				  LookAndFeel originalLaf = UIManager.getLookAndFeel();
				  try {
					  UIManager.setLookAndFeel (UIManager.getSystemLookAndFeelClassName());
				  } catch (Exception eLF) {
						  eLF.printStackTrace();
				  }	
	    	
	    	      //File selection dialogue with extension filter
				  imageChooser.updateUI();
				  imageChooser.setFileSelectionMode (JFileChooser.FILES_ONLY);
				  imageChooser.setFileFilter (imgFilter);
	    	      int iSalida = imageChooser.showOpenDialog (getWindow());
	    	      
	    	      //Flick the L&F back to the default
	    		  try {
	    			  UIManager.setLookAndFeel(originalLaf);
	    		  } catch (UnsupportedLookAndFeelException eLF) {
	    			  	  eLF.printStackTrace();
	    		  }
	    	      
	    	      //An image file has been selected => insert it using <IMG> tag
	              if (iSalida == JFileChooser.APPROVE_OPTION) {
	            	 insertImg (imageChooser.getSelectedFile().getAbsolutePath(), iSelStart,
	            			    iSelEnd);
	              }	            	 
	              
	    //Table insert button
	    } else if (e.getActionCommand().equals ("insertarTabla")) {
	    	      int iRows = 0,  		//number of rows
	    	          iCols = 0,    	//number of columns 
	    	          iColWidth = 0;   	//column's width
	    	      
	    	      //Displays the dialogue requesting for the new table's parameters
	    	      if (tableConfig.muestraCampoConfirmar (getWindow())) {
	    	    		   
	    	         //Gets chosen number of rows and columns respectively
	    	         iRows = Integer.valueOf (tableConfig.getFields()[0]); 	         
	    	         iCols = Integer.valueOf (tableConfig.getFields()[1]);
	    	    	        	    	    	     
	    	         //Sets out the column width according to the text area's width (effectively stretching the table)
		    	     iColWidth = textArea.getWidth() / iCols;    	             	    	                  
	    	      }	    	         
	    	      
	    	      if (iColWidth != 0)
	    	         insertTable (iRows, iCols, iColWidth);	    	    
	    		
	    //Bulleted list insert button
	    } else if (e.getActionCommand().equals ("insertarVineta")) {
	    		  
	    	      //Adds/removes the necessary list tags
	    	      insertList ("ul", iSelStart, iSelEnd);
	    	      	    	      
	    	      //Resets the numbered list insert button in case it was on
	    	      if (buttons.containsKey("insertarNumeracion") && buttons.get("insertarNumeracion").isSelected()) {
		    	     buttons.get("insertarNumeracion").setSelected (false);
		    	  }
	    	      
	    	      //Checks if the selected text and the caret are both within the list
	    	      if (buttons.containsKey("insertarVineta")) {
	    	         Element raiz = getNearestParent (document, textArea.getCaretPosition(), "ul body");
	    	         if (raiz.getName().equals("ul") && ((iSelEnd <= raiz.getEndOffset()) && (iSelStart >= raiz.getStartOffset())))
		    		    buttons.get("insertarVineta").setSelected (true);
		    	     else buttons.get("insertarVineta").setSelected (false);
	    	      }
	    
	    //Numbered list insert button
	    } else if (e.getActionCommand().equals ("insertarNumeracion")) {
	    	      
	    		  //Adds/removes the necessary list tags
	    	      insertList ("ol", iSelStart, iSelEnd);
	    	      
	    	      //Resets the numbered list insert button in case it was on 
	    	      if (buttons.containsKey("insertarVineta") && buttons.get("insertarVineta").isSelected()) {
			    	 buttons.get("insertarVineta").setSelected (false);
			      }
	    	      
	    	      //Checks if the selected text and the caret are both within the list
	    	      if (buttons.containsKey("insertarNumeracion")) {
	    	         Element root = getNearestParent (document, textArea.getCaretPosition(), "ol body");
	    	         if (root.getName().equals("ol") && ((iSelEnd <= root.getEndOffset()) && (iSelStart >= root.getStartOffset())))
		    		    buttons.get("insertarNumeracion").setSelected (true);
		    	     else buttons.get("insertarNumeracion").setSelected (false);
	    	      }	    	      
	    	      	    	      
	    //Left indent button 
	    } else if (e.getActionCommand().equals ("sangria")) {
	    	      final float fNewIndent;		//new value for left indentation
	    	      MutableAttributeSet attrib;	//paragraph left indent attribute, either new or existing
	    	      
	    	      //If the text is already left-indented, remove that indentation
	    	      attrib = (MutableAttributeSet) document.getParagraphElement(iSelStart).getAttributes();
	    	      if (StyleConstants.getLeftIndent (attrib) == 0.0F)
	    	    	 fNewIndent = Global.INDENTATION;
	    	      else fNewIndent = 0.0F;
	    	      
	    	      //Applies the established indentation value
		    	  attrib = new SimpleAttributeSet();
	              StyleConstants.setLeftIndent (attrib, fNewIndent);
	              textArea.setParagraphAttributes (attrib, false);
	    
	    //Line spacing button (considered as the space BELOW each line)
	    } else if (e.getActionCommand().equals ("interlineado")) {
	    	      final float fNewLineSpacing;	//new value for line spacing
	    	      MutableAttributeSet attrib;	//paragraph line spacing attribute, either new or existing 
	    	      
	    	      //If lines already have a space below, remove that space
	    	      attrib = (MutableAttributeSet) document.getParagraphElement(iSelStart).getAttributes();
	    	      if (StyleConstants.getSpaceBelow (attrib) == 0.0F)
	    	         fNewLineSpacing = Global.LINE_SPACING; 
	    	      else fNewLineSpacing = 0.0F;
	    	      
	    	      //Applies the established spacing.
	    	      attrib = new SimpleAttributeSet();
                  StyleConstants.setSpaceBelow (attrib, fNewLineSpacing);
                  textArea.setParagraphAttributes (attrib, false);
	    	     
        //Copy/Paste button
	    } else if (e.getActionCommand().equals ("copiarOPegar")) {
	    	      if (bCopy) //a copy has previouly been made => next thing now is to paste
	    	    	 paste ();
	    	    	     	      
	    	      //The copied text has already been pasted => next thing now is to copy    
	    	      else copy();
	    	      
	    //Undo button
	    } else if (e.getActionCommand().equals ("deshacer")) {
	    	      undo(); 
	    	      
	    //Inline pictograms button
	    } else if (e.getActionCommand().equals ("pictogramas")) {
	    	      bPictosOn = !bPictosOn;
	    	      buttons.get("pictogramas").setSelected();
	    	      try {
	    	          if (!bPictosOn) 
	    	    	     removePictos (document, textArea, false);
	    	    	  else textArea.setCaretPosition (insertPictos(getBody(document).getStartOffset(), iCaretPos) + iCaretPos);	    	      
	    	      } catch (Exception ePictos){
	    	    	      System.out.println ("Error al insertar/borrar pictogramas: " +ePictos);
	    	    	      ePictos.printStackTrace();
	    	      }    	 
	    
	    //Background-foreground colour pattern button
	    } else if (e.getActionCommand().equals ("cambiarColorFondo")) {
	    	      iCurrBgColour = (iCurrBgColour + 1) % 3;  //only 3 possible patterns
	    	      Color fgColour = fgColours.get (bgColours[iCurrBgColour]);
	    	      String sColorHex = Integer.toHexString (fgColour.getRGB() & 0x00ffffff);
	    	      
	    	      //Changes background colour. Also the caret's colour is set to the text's default
	    	      textArea.setBackground (bgColours[iCurrBgColour]);
	    	      textArea.setCaretColor (fgColour);
	    	      
	    	      //Changes the text's default colour to that according to background
	    	      globalStyles.addRule ("body {color: #" +sColorHex+ "}");
	    	      globalStyles.addRule ("table {border-color: #" +sColorHex+ "}");
	    	      globalStyles.addRule ("td {border-color: #" +sColorHex+ "}");
	    	      globalStyles.addRule ("li {color: #" +sColorHex+ "}");
	    	      
	    	      //If PBDQ is on, change previous PBDQ colours to those arranged for the current background
	    	      if (bPBDQOn) 
	    	         pbdqParser (true, getBody(document).getStartOffset(), document.getLength() - 1, 
	    	        		     document, textArea);
	    	      
	    	      //If pictograms on, change the foreground colour of any label to match the document's
	    	      if (bPictosOn && (iCurrBgColour > 0)) {
	    	    	 HTMLDocument.Iterator iterator;  
	    	    	 
	    	    	 //For all IMG tags...
	    	       	 for (iterator = document.getIterator(HTML.Tag.IMG); iterator.isValid(); 
	    	       	      iterator.next())
	    	       		 
	    	       		 //Consider only those with an ID attribute (pictogram image tags)
	    	       		 if (iterator.getAttributes().isDefined (HTML.Attribute.ID))
	    	    	   	    ((JLabel) StyleConstants.getComponent (iterator.getAttributes()))
	    	  			                            .setForeground (fgColour);    	  		 	   
	    	      }
	    	      
	    //Inline PBDQ button
	    } else if (e.getActionCommand().equals ("resalta")) {
	    	      bPBDQOn = !bPBDQOn;
	    	      buttons.get("resalta").setSelected();
	    	      pbdqParser (bPBDQOn, getBody(document).getStartOffset(), document.getLength() - 1, 
	    	    		      document, textArea);
	    	      if (bPBDQOn) //muestra leyenda al activar función
	    	    	 PBDQkey.display();
	    	      else PBDQkey.close();
	    	      textArea.setCaretPosition (iCaretPos);
	    	      
	    //Virtual keypad button: displays a pop-up menu with all available keypads
	    } else if (e.getActionCommand().equals ("tareas")) {
  	              if (!bVKeypadOn) //Displays virtual keypad menu on first click
  	            	 vKeypadMenu.show ((JButton) e.getSource(), 20, 20);
  	              else {           //On switching off, closes the virtual keypad panel.
  	                   getBottomPanel().setVisible (false);
  	    	           bVKeypadOn = !bVKeypadOn;
  	  	               buttons.get("tareas").setSelected();
  	              } 	            	 
  	              
  	    //A button within the currently loaded virtual keypad has been pressed
        } else if (e.getActionCommand().startsWith ("tecladoVirtualInsert")) {
        	      JButton vKeypadButton;	 //current virtual keypad's button
        	      SimpleAttributeSet attrib; //text attributes to be applied  
        	      String sButtonText;		 //text defined for the keypad's button        	             
        	      int iTextSize,			 //text's size
        	          iCaretPosAfter;		 //caret's position after inserting text       	      
        	              	      
        	      vKeypadButton = (JButton) e.getSource();
        	      attrib = null;
        	      try {
	    	    	  //Button belongs to a type-2 virtual keypad? => highlight inserted words
	    	          if (e.getActionCommand().endsWith ("2")) { 
	    	         	 attrib = new SimpleAttributeSet();
	    	    	     if (vKeypadButton.getMnemonic() == -1) 
	    	    	        StyleConstants.setForeground (attrib, Global.WRONG);
	    	    	     else StyleConstants.setForeground (attrib, Global.RIGHT);
	    	          }
	    	          
	    	          //Gets text to be inserted and suffixes it with a space.
	    	          sButtonText = vKeypadButton.getToolTipText().replaceAll (" ", "\u00A0");
	    	          sButtonText = sButtonText+ "\u00A0";
	    	          iTextSize = sButtonText.length();
	    	              	              	          
	    	          //If it's not the beggining of a paragraph and there is no previous image or space, 
	    	          //prefix the text with a space.
	    	          if ((document.getParagraphElement(iSelStart).getStartOffset() != iSelStart) && 
	    	        	  (document.getCharacterElement(iSelStart - 1).getName().equals("img") || 
	    	        	   !textArea.getText(iSelStart - 1,1).equals("\u00A0"))) { 
	    	          	 sButtonText = "\u00A0" +sButtonText;
	    	        	 iTextSize++;
	    	          }
	    	          
	    	          //Attributes at the caret's current position are conserved, except for those representing
	    	          //a pictogram already inserted.
	    	          preventInsPicto (textArea.getInputAttributes());
	    	          
	    	          //Inserts the text. If the keypad is of type 2, it highlights the text accordingly.
	    	          textArea.replaceSelection (sButtonText);
	    	          if (attrib != null)
	    	          	 document.setCharacterAttributes (iSelStart, iTextSize, attrib, false);
	    	          
	    	          //If any accessibility feature is on, make the corresponding changes.
	    	          if (bPBDQOn)
	    	        	 pbdqParser (true, iSelStart, iSelStart + sButtonText.length(), 
	    	        			     document, textArea);
	    	          if (bPictosOn) {
	    	        	 iCaretPosAfter = textArea.getSelectionEnd(); 
	    	        	 textArea.setCaretPosition (insertPictos (iSelStart, iCaretPosAfter) + iCaretPosAfter);
	    	          } 	
	    	       } catch (Exception eTexto) {
				           System.out.println ("Problemas al insertar texto: " +eTexto);
					       eTexto.printStackTrace();
			       }
		
	    //Displays type-1 virtual keypad
	    } else if (e.getActionCommand().equals ("tecladoVirtual1")) {
	    	      showKeypad (((JMenuItem)e.getSource()).getText()+ ".xml", 
	    	    		     "tecladoVirtualInsert1", true);			      
	    	      
	    //Displays type-2  virtual keypad
	    } else if (e.getActionCommand().equals ("tecladoVirtual2")) {
	    		  showKeypad (((JMenuItem)e.getSource()).getText()+ ".xml", 
		                     "tecladoVirtualInsert2", false);		    	      	
	    }
			
		//Has a text-related action taken place?
		if (!e.getActionCommand().isEmpty() && !PBDQkey.isVisible()) {
	       textArea.grabFocus();
	       bGrabFocus = false;
		} else bGrabFocus = true;
	}
	
	/**
	 * Takes care of changes in the document model at the press of a button: 
	 * 1.Deletes an empty item list on pressing the ENTER key. 
	 * 2.Manages backspace and delete key events factoring the surrounding text in
	 * 3.Inserts a pictogram once the space bar has been pressed (if pictograms on).
	 * 4.Changes the colour of the "p", "b", "d" or "q" characters if PBDQ is on and the corresponding
	 * key has been pressed.
	 */
	public void keyPressed(KeyEvent e) {
		char character;				  //character corresponding to pressed key
		int iStartSec, iEndSec,		  //section's starting and ending offsets
		    iStartSel, iEndSel,	      //selection's starting and ending offsets		
		    iStartPar, iEndPar,		  //paragraph's starting and ending offsets
		    iStartDoc, iEndDoc,		  //document's starting and ending offsets 
		    iCaretPos,				  //caret's current position	
		    iKeyCode;				  //number code for the pressed key
		boolean bEmptyElement,		  //does the element contain no non-blank text?
		        bPartialSel,	  	  //does the seletion include any structure partially?	
		        bHasImage;		  	  //does the current paragraph contain any image?
		String sTITLEvalue,			  //value for the TITLE attribute
		       sSectionTag,			  //current section's tag name
		       sPriorTag,	  	  	  //previous section's tag name
		       sNextTag,	  	  	  //following section's tag name	
		       sHTML;				  //HTML string
		Element section, 			  //current paragraph's parent element; can be LI/TD/BODY
        		blockRoot,			  //root element of current list or table
        		currParagraph;	  	  //paragraph including caret's current position
		MutableAttributeSet attribSPAN,//attributes of <SPAN> tag 
							attribFONT,//attributes of <FONT> tag 
                    		attribIns; //current insert attributes
		 
		//Initialisations
		iKeyCode = e.getKeyCode();
		character = Character.toLowerCase (e.getKeyChar());
		iCaretPos = textArea.getCaretPosition();
		iStartDoc = getBody(document).getStartOffset();
		iEndDoc = getBody(document).getEndOffset();
		iStartSel = textArea.getSelectionStart();
		iEndSel = textArea.getSelectionEnd();
		currParagraph = document.getParagraphElement(iStartSel);
		iStartPar = currParagraph.getStartOffset();
		iEndPar = currParagraph.getEndOffset();
        section = currParagraph.getParentElement();
        sSectionTag = section.getName();
        sPriorTag = document.getParagraphElement(iCaretPos - 1).getParentElement().getName();
        sNextTag = document.getParagraphElement(iCaretPos + 1).getParentElement().getName();
        bHasImage = (document.getElement (currParagraph, StyleConstants.NameAttribute, HTML.Tag.IMG) != null);
        attribIns = textArea.getInputAttributes();
        
        //Cleans input attributes of any remnants of pictograms. 
        if ((character != KeyEvent.VK_BACK_SPACE) &&
  		    (character != KeyEvent.VK_DELETE) &&
  		    (character != KeyEvent.CHAR_UNDEFINED) &&
  		    (iStartSel == iEndSel)) {
  	       preventInsPicto (attribIns); 	       
        }
  	         
        //Not a list or table element
        if (sSectionTag.equals ("body")) {
           section = currParagraph;
           sSectionTag = "p";   
        }
        
        iStartSec = section.getStartOffset();
  	    iEndSec = section.getEndOffset();
  	     	  	    
		try {
			bEmptyElement = textArea.getText(iStartSec, iEndSec - iStartSec).trim()
                                     .replaceAll ("[\\p{Z}\\s]", "").isEmpty();
			
			//Was the ENTER key pressed at an existing list item?
            if ((iKeyCode == KeyEvent.VK_ENTER) && sSectionTag.equals ("li")) {
        	 
               //Caret at non-empty list item? => add a new item below
               if (bHasImage || !bEmptyElement) {
            	  bNewListItem = true;    	        	  
            	              				 
                //List item empty => remove it
                } else {
                       //Purges ENTER's keypress event
                	   e.consume();
                	                   	                     	                    	                  	   
                	   //Inserts new paragraph and partition any existing list if necessary
                	   editorKit.insertHTML (document, iEndSec, sEmptyParagraph, 3, 0, HTML.Tag.P);
                	   
                	   //Deletes the empty list item
                	   document.remove (iStartSec, iEndSec - iStartSec);
                } 
           
           //Selection and replacing selection when a character with representation has been pressed
           } else if ((iStartSel != iEndSel) && (character != KeyEvent.CHAR_UNDEFINED)) {
        	         
        	         //No special key combination exists (like Ctrl+V)?   
        	         if (e.getModifiersEx() == 0) {
        	   		    //Focuses on the tags at the selection's start and end (instead of previous and next)
      	                sPriorTag = document.getParagraphElement(iStartSel).getParentElement().getName();
                        sNextTag = document.getParagraphElement(iEndSel).getParentElement().getName();
                     
                        //Determines the selected structure's root element
                        if (sSectionTag.equals("td"))
                    	   blockRoot = section.getParentElement().getParentElement();
                        else if (sSectionTag.equals("li"))
                    	        blockRoot = section.getParentElement();
                             else blockRoot = section; 
                     
                        //Does the selection cover the whole structure?
                        bPartialSel = ((iStartSel > blockRoot.getStartOffset()) || (iEndSel < (blockRoot.getEndOffset() - 1)));
                     
                        //The selection crosses the boundary between different structures (e.g. a list and a paragraph)
                        //=> discard selection
                        if (!sPriorTag.equals(sNextTag)) {
                  	       e.consume();
                  	       bPurgeKeyPress = true;
                  	    
                  	       //The selection includes the whole structure where the caret is
                  	       //=> delete the structure
                  	       if (!bPartialSel)
                  	          if (blockRoot.getEndOffset() == iEndDoc)
                  	             document.remove (blockRoot.getStartOffset(), blockRoot.getEndOffset() - blockRoot.getStartOffset() - 1);
                  	          else document.remove (blockRoot.getStartOffset(), blockRoot.getEndOffset() - blockRoot.getStartOffset());
                  	    
                  	       //If not a structure but a paragraph (except last one) => remove it
                  	       else if (sSectionTag.equals("p") && (iStartSel <= iStartPar) && (iEndSel >= iEndPar) &&
                  	    		    (iEndSec != iEndDoc))
                  	    	       document.remove (iStartPar, iEndPar - iStartPar);
                  	    
                  	       //If backspace was pressed => move the caret back one position.
                  	       if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos > getBody(document).getStartOffset()))
    	        	          textArea.setCaretPosition (iCaretPos - 1);
                  	                         
                  	    //Selection includes various elements within same structure but in different paragraphs
                  	    //=> discard selection
                        } else if (((blockRoot.getElementIndex(iStartSel) != blockRoot.getElementIndex(iEndSel)) ||
                    	   	        (section.getParentElement().getElementIndex(iStartSel) != section.getParentElement().getElementIndex(iEndSel))) &&
                    	           bPartialSel && !sSectionTag.equals("p"))  {
                    	          e.consume();
                    	          bPurgeKeyPress = true;
                    	          if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos > iStartDoc))
            	        	         textArea.setCaretPosition (iCaretPos - 1);                    	          
                    
                        //Selection includes the structure at the caret's position entirely => remove it	       
                        } else if ((!bPartialSel) && !sSectionTag.equals("p")) {
                    	          e.consume();
              	                  document.remove (blockRoot.getStartOffset(), blockRoot.getEndOffset() - blockRoot.getStartOffset());              	               
              	                  if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos > iStartDoc))
      	        	                 textArea.setCaretPosition (iCaretPos - 1);

              	        //Selected paragraph is empty and the last one in document => do not remove           
                        } else if (sSectionTag.equals("p") && (iEndSec == iEndDoc) && 
                    	   	      bEmptyElement && !bHasImage) {
                    	          e.consume();  
                    	          if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos > iStartDoc))
       	        	                 textArea.setCaretPosition (iCaretPos - 1);
                        }
                     
                     //Key combination: character + mousebutton => avoid default replacement event   
        	         } else if (KeyEvent.getModifiersExText(e.getModifiersEx()).contains("Button")){
        	        	       e.consume();
        	        	       bPurgeKeyPress = true;
        	         }
           //A line brake about to be deleted using backspace from the start of a line and without selection         
           } else if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos > iStartDoc) && 
        		      (iCaretPos == currParagraph.getStartOffset()) && (iStartSel == iEndSel) &&
                      (textArea.getText(iCaretPos - 1, 1).equals("\n"))) {
        	         
        	         //There is a table cell before => inside a table => do not delete cells
                     if ((sPriorTag.equals("td") && sSectionTag.equals("p")) || 
                    	 (sSectionTag.equals("td") && (iCaretPos == section.getStartOffset()))) {
        	            e.consume(); 
        	            if ((sSectionTag.equals("p") && bEmptyElement && !bHasImage)
        	                && (iEndSec != iEndDoc))
        	        	   document.remove (iStartSec, iEndSec - iStartSec);
        	            if (iCaretPos > iStartDoc)
        	               textArea.setCaretPosition (iCaretPos - 1);  
        	         
	        	     //If current paragraph is inside a list...
        	         } else if (sSectionTag.equals("li")) {
        	        	       e.consume();
                                                                                       
                               //Item not first one => Merges item with previous one and moves line break to end 
                               if (section.getParentElement().getElementIndex(iCaretPos) > 0) {
                            	  editorKit.runCommand("delete-previous");
                                  document.insertString (iCaretPos - 1, "\n", textArea.getInputAttributes()); 
                            	  textArea.setCaretPosition (iCaretPos - 1); 
                            	  editorKit.runCommand("delete-next");
                        	
                               //Item is list's first: revert item to paragraph and merge it with previous one
                               } else {
                            	      insertList (section.getParentElement().getName(), iCaretPos, iCaretPos);
                            	      if (!sPriorTag.equals("td"))
                            	    	  editorKit.runCommand("delete-previous");
                               }
                                      	                	         
        	         //If current paragraph is outside list but inmediately after it
        	         } else if ((sSectionTag.equals("p") || sSectionTag.equals("td")) 
        	        		    && sPriorTag.equals("li")) {
        	        	 	   HashMap<String, SimpleAttributeSet> pictoList = new HashMap<String, SimpleAttributeSet>();
        	        	 	           	        	       
        	        	 	   e.consume();  
        	        	              	        	 
        	        	       //Backs up the paragraph's contents
        	        	       if (bHasImage || !bEmptyElement) {
        	        	          sHTML = extractHTML (currParagraph, new StringWriter());        
                                  sHTML = sHTML.substring(sHTML.indexOf(">") + 1, sHTML.length());
                                  sHTML = sHTML.substring(0, sHTML.lastIndexOf("</p>"));
        	        	       } else sHTML = ""; 
        	        	       
        	        	       //Backs up the attributes for every affected pictogram
                  		       pictoList = backUpPictos (iStartPar, iEndPar);
                               
                               //Deletes the current paragraph if it is not the document's last
                               if (iEndSec != iEndDoc) {
                            	  if  ((iEndPar != iEndSec) && sSectionTag.equals("td"))
                                      document.remove (iStartPar, iEndPar - iStartPar);
                            	  else if (!sSectionTag.equals("td"))
                            		      document.remove (iStartPar, iEndPar - iStartPar); 
                               
                               //Last paragraph => remove it but be careful with EOF
                               } else document.remove (iStartPar, iEndPar - iStartPar - 1);
                               
                               //Adds the backed up contents to those of the last list item
                               document.insertBeforeEnd (document.getParagraphElement(iCaretPos - 1), sHTML);
                                                              
                               //Fixes pictograms corrupted on inserting the new list item.
                               fixPictos (pictoList);                               
                               
                               //Resets the caret position (to give impression nothing has changed)
                			   textArea.setCaretPosition (iCaretPos - 1);
                               
                     //If paragraph is the document's last and is empty => do not remove
        	         } else if ((iEndSec == iEndDoc) && bEmptyElement && !bHasImage) {
        	        	       e.consume(); 
                               textArea.setCaretPosition (iCaretPos - 1);
                     }
                             
           //A line brake about to be deleted using delete at the end of a line without selection     
           } else if ((iKeyCode == KeyEvent.VK_DELETE) && (iStartSel == iEndSel) &&
        		      (iCaretPos == (currParagraph.getEndOffset() - 1)) &&
                      (textArea.getText(iCaretPos, 1).equals("\n"))) {
        	         
        	         //There is a table cell after => inside a table => do not delete cells 
  	                 if ((sNextTag.equals("td") && sSectionTag.equals ("p") &&
  	                     (!bEmptyElement || bHasImage || (iEndSec == iEndDoc))) 
  	                     || (sSectionTag.equals("td") && (iCaretPos == (section.getEndOffset() - 1)))) {
  	        	        e.consume(); 
  	        	        
        	         //If paragraph inside a list => delete current paragraph move back to previous one
  	                 } else if (sSectionTag.equals("li")) {
  	                	       e.consume();
  	                	       Element listRoot = section.getParentElement();
  	                	       
  	                	       //Item not first one => Merges item with next one and moves line break to end 
                               if (listRoot.getElementIndex(iCaretPos) < (listRoot.getElementCount() - 1)) {
                            	  editorKit.runCommand ("delete-next");
                                  document.insertString (iCaretPos, "\n", textArea.getInputAttributes());
                                  editorKit.runCommand ("delete-previous");   
                            	   
                               //Item is list's last
                               } else {
                            	      HashMap<String, SimpleAttributeSet> pictoList = new HashMap<String, SimpleAttributeSet>();
                            	                                  	                                  	      
                            	      //Get parameters of the paragraph after the list
                            	      currParagraph = document.getParagraphElement (iCaretPos + 1);
                            	      iStartPar = currParagraph.getStartOffset();
                            	      iEndPar = currParagraph.getEndOffset();
                            	      
                            	      //Backs up that paragraph's contents
            	        	          sHTML = extractHTML (currParagraph, new StringWriter());        
            	        	          sHTML = sHTML.substring(sHTML.indexOf(">") + 1, sHTML.length());
            	        	          sHTML = sHTML.substring(0, sHTML.lastIndexOf("</p>"));
                                   
            	        	          //Backs up the attributes for every affected pictogram
                         		      pictoList = backUpPictos (iStartPar, iEndPar);
            	        	          
                                      //Deletes that paragraph if it is not the document's last
                                      if (iEndPar != getNearestParent (document, iCaretPos + 1, "td body").getEndOffset())
                                         document.remove (iStartPar, iEndPar - iStartPar);
                                   
                                      //It is the last paragraph => remove it but be careful with EOF
                                      else document.remove (iStartPar, iEndPar - iStartPar - 1); 
                                      
                                      //Adds the backed up contents to those of the last list item
                                      document.insertBeforeEnd (document.getParagraphElement(iCaretPos), sHTML);
                                                                            
                                      //Fixes pictograms corrupted on inserting the new list item.
                                      fixPictos (pictoList);
                                      
                                      //Resets the caret position (to give impression nothing has changed)
                       			      textArea.setCaretPosition (iCaretPos);
                               }          
                        
                     //Paragraph outside list but immediately before => adds first list item to end of paragraph
                     } else if ((sSectionTag.equals("p") || sSectionTag.equals("td")) 
        	        		    && sNextTag.equals("li")) {
    	                       e.consume();
        	        	       insertList (section.getParentElement().getName(), iCaretPos + 1, iCaretPos + 1);
        	        	       editorKit.runCommand ("delete-previous");   	                 	             
        	         }
  	                 
  	       //If paragraph is the document's last and is empty => do not remove          
           } else if (((iKeyCode == KeyEvent.VK_DELETE) || (iKeyCode == KeyEvent.VK_BACK_SPACE))  
        		      && (iEndSec == iEndDoc) 
        		      && bEmptyElement && !bHasImage) {
        	         e.consume();
        	         if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos > iStartDoc))
   	        	        textArea.setCaretPosition (iCaretPos - 1); 
        	         
           //If backspace pressed at document's start and no list in sight => do not remove
           } else if ((iKeyCode == KeyEvent.VK_BACK_SPACE) && (iCaretPos == iStartDoc))
        		     if (!sSectionTag.equals("li")) 
        	            e.consume();
                     
                     //Inside an item list => revert to paragraph
        		     else insertList (section.getParentElement().getName(), iCaretPos, iCaretPos);
            
		} catch (Exception eTexto) {
 	            System.out.println ("Problemas al eliminar/insertar texto: " +eTexto);
                eTexto.printStackTrace();
        } 
		
		//Insert pictogram if space/tab/enter/'/-/_ is pressed after a word and pictograms are on.
		if (((iKeyCode == KeyEvent.VK_SPACE) || (iKeyCode == KeyEvent.VK_ENTER) || 
		     (iKeyCode == KeyEvent.VK_TAB) || (character == '\'') || (character == '-') ||
		     (character == '_')) && 
			(StyleConstants.getComponent (document.getCharacterElement(iStartSel - 1).getAttributes()) == null) &&
			(iStartSel > iStartPar) && bPictosOn) 
		   insertPicto (null, iCaretPos);		   
			    
		//Change the colour of character after p/b/d/q is pressed and PBDQ on. 
		if (bPBDQOn && ((character == 'p') || (character == 'b') || 
	                    (character == 'd') || (character == 'q'))) {
		   
		   //Prevents character insertion due to key press
		   e.consume();
		   bPurgeKeyPress = true;
		   
		   //Adds the corresponding character replacing previous text if necessary
		   replaceString (Character.toString (e.getKeyChar()), iStartSel, iEndSel, 
				       document, textArea);
		   		  		   
		   //Gets insert attributes
		   attribIns = textArea.getInputAttributes();
		   
		   //Filters out PBDQ attributes coming from a previous PBDQ character
		   if (attribIns.isDefined (HTML.Tag.SPAN)) {
	          attribIns.removeAttribute (CSS.Attribute.COLOR);
	          attribSPAN = (SimpleAttributeSet) attribIns.getAttribute (HTML.Tag.SPAN);
	          sTITLEvalue = (String) attribSPAN.getAttribute(HTML.Attribute.TITLE);
	          if (!sTITLEvalue.equals("defecto"))
	             StyleConstants.setForeground (attribIns, new Color (Integer.valueOf (sTITLEvalue)));
	          attribIns.removeAttribute (HTML.Tag.SPAN);
		   }
		   
		   //Backs up the current colour of the character and represents it as a TITLE attribute
		   if (!attribIns.isDefined (CSS.Attribute.COLOR)) { //default colour?
	    	  if (!attribIns.isDefined(HTML.Tag.FONT))
	    		 sTITLEvalue = "defecto";
	    	  else {
	    		   attribFONT = (SimpleAttributeSet) attribIns.getAttribute (HTML.Tag.FONT);
	    		   if (!attribFONT.isDefined (HTML.Attribute.COLOR))
	    			  sTITLEvalue = "defecto";
	    		   else sTITLEvalue = String.valueOf (StyleConstants.getForeground (attribIns).getRGB());
	    	  }	    	   
	       } else sTITLEvalue = String.valueOf (StyleConstants.getForeground (attribIns).getRGB());   
	
		   //The TITLE attribute will belong to a SPAN tag written inside the necessary FONT tag  
	       attribIns = new SimpleAttributeSet();
	       attribSPAN = new SimpleAttributeSet();
 		   attribSPAN.addAttribute (HTML.Attribute.TITLE, sTITLEvalue);
      	   attribIns.addAttribute (HTML.Tag.SPAN, attribSPAN);
      	   document.setCharacterAttributes (iStartSel, 1, attribIns, false);
      	   
      	   //Sets the character's colour to the predefined one for PBDQ
		   setFontColour (iStartSel, iStartSel + 1, mapPBDQ.get(character)[iCurrBgColour], document, true);   	   
      	
		//Prevents the PBDQ colour spreading to subsequent characters  
		} else if (bPBDQOn) 
      	          preventInsPBDQ (textArea.getInputAttributes());     		
	}    

	/**
	 * Performs various changes on the document model: 
	 * 1.Turns any present implied paragraph into real one,
	 * 2.Implements the insertion of a new list item on pressing the enter button at a previous one 
	 */
	public void keyReleased (KeyEvent e) {
		int iCaretPos,			  //caret's current position	
		    iDocStart,        	  //character offset the document starts from
		    iStartPar, iEndPar;	  //starting and ending offsets of current paragraph	
		boolean bNewItemEmpty;	  //el nuevo item de lista está vacío?
		String sHTMLListItem;	  //HTML code representing a new list item 
		BlockElement currParagraph, 	//paragraph at caret's current position 
		             parentParagraph;	//parent element of current paragraph (either TD or BODY)
		SimpleAttributeSet atribPreviousItem;	  //attributes at the end of the previous list item
		HashMap<String,SimpleAttributeSet> pictoList = new HashMap<String,SimpleAttributeSet>();
	
	    iDocStart = getBody(document).getStartOffset();						
		iCaretPos = textArea.getCaretPosition();
		bNewItemEmpty = false;
		atribPreviousItem = null;
		try {
			
			//Make sure caret's current position is not before the beginning of the document
			if (iCaretPos < iDocStart) {
			   textArea.setCaretPosition (iDocStart);
			   iCaretPos = iDocStart;			   
			}
			
			//Turns any p-implied paragraph (e.g. when starting a new document) to a real one
			SimpleAttributeSet atribParrafo = new SimpleAttributeSet(textArea.getParagraphAttributes());
			if (atribParrafo.getAttribute (Style.NameAttribute).toString().equals("p-implied")) {
			   atribParrafo.addAttribute (AttributeSet.NameAttribute, HTML.Tag.P);
			   StyleConstants.setSpaceAbove (atribParrafo, 0f);
			   textArea.setParagraphAttributes (atribParrafo, true);
			}
				
			//ENTER key has been pressed at a previous non-empty list item => insert new item
			if (bNewListItem) {
			   
			   //Set up new item from the already partitioned list (taken care of by the key pressed event)
			   currParagraph = (BlockElement) document.getParagraphElement (iCaretPos);
			   iStartPar = currParagraph.getStartOffset();
			   iEndPar = currParagraph.getEndOffset();
      	       sHTMLListItem = extractHTML (currParagraph, new StringWriter());
               sHTMLListItem = "<li>" +sHTMLListItem+ "</li>";
               
               //Insert an empty paragraph if there is no text after the new item
			   parentParagraph = (BlockElement) getNearestParent (document, iCaretPos, "td body");
         	   if (iEndPar == parentParagraph.getEndOffset())
                  sHTMLListItem = sHTMLListItem + sEmptyParagraph;
         	   
         	   //There is no text coming from previous list item => retrieve attributes from previous line 
               if (textArea.getText(iCaretPos, iEndPar - iCaretPos).trim().isEmpty()) {
                  atribPreviousItem = new SimpleAttributeSet (textArea.getInputAttributes().copyAttributes());
                  bNewItemEmpty = true;
               }
         	 
               //Back up the attributes for every affected pictogram
		       pictoList = backUpPictos(iStartPar, iEndPar);

               //Replaces the paragraph with the newly built list item        	   
			   document.remove (iStartPar, iEndPar - iStartPar);
			   editorKit.insertHTML (document, iStartPar, sHTMLListItem, 
					                 2, 0, HTML.Tag.LI);
			   		   
			   //Fixes pictograms corrupted during the insertion of the new list item.
			   fixPictos (pictoList);
			   
			   //Restores the caret to its position before inserting the new item.
			   textArea.setCaretPosition (iCaretPos);
			   
			   //Re-applies attributes from the previous list item
			   if (bNewItemEmpty)
				  textArea.setCharacterAttributes (atribPreviousItem, false); 
			   		   
			   bNewListItem = false;			   
			}			    
		} catch (Exception eTexto) {
			    System.out.println ("Problemas al insertar texto: " +eTexto);
			    eTexto.printStackTrace();
		}	
    }
	
	/**
	 * Updates the status of all toolbar buttons according to the format of the selected text or that
	 * text where the caret is currently positioned. It also enables the undo button on any text changes.
	 * Finally, the copy/paste button is enabled once some text has been selected.
	 * @param iCaretPos Caret's current position.
	 */
	public void updateUI (int iCaretPos) {
		String sCurrentFont,			//font type present at the caret's position
	       	   sCurrentList;			//list type present at the caret's position
		JComboBox dropDownMenu;			//drop-down meny with all font types
		AttributeSet attrib;			//attributes to be probed for format information
		int iStartSel,					//starting offset of selected text
			iEndSel,					//ending offset of selected text
			iDocStart;					//character offset the document starts from
		boolean bInsideList;			//is the caret inside a list?		        
	
		//Avoid any updating while the document is being loaded
		if (document != null) {
	
			iDocStart = getBody(document).getStartOffset();
			//Ensures the caret is always "inside" the body element
			if (iCaretPos < getBody(document).getStartOffset())
			   textArea.setCaretPosition (iDocStart);
	
			//Finds out the current selection range and corresponding attributes
			iStartSel = textArea.getSelectionStart();
			iEndSel = textArea.getSelectionEnd();
			if (iEndSel - iStartSel == 0) //no selection? => take attributes of last char 
			   if (iCaretPos > iDocStart) 
				  attrib = document.getCharacterElement (iCaretPos - 1).getAttributes();
			   else attrib = textArea.getCharacterAttributes();
			else attrib = document.getCharacterElement (iStartSel).getAttributes();
	
			if (buttons.containsKey("fuente")) {
			   //Retrieves the current font
			   sCurrentFont =  StyleConstants.getFontFamily (attrib);
	
			   //Retrieves the drop-down menu of font types (to change currently selected font) 
			   dropDownMenu = (JComboBox) buttons.get("fuente").getComponent();
	      
			   //Changes selected font if necessary
			   dropDownMenu.removeActionListener (this); //avoids unnecessarily triggering event
			   if (sCurrentFont.equals ("Monospaced"))   //Swing's default font? => change to editor's default 
				  dropDownMenu.setSelectedItem (Global.FONT_FAMILY); 	
			   else dropDownMenu.setSelectedItem (sCurrentFont);
			   dropDownMenu.addActionListener (this);
			}
	   
		   //Manages the representation of the bold, italic and underline formats
		   if (buttons.containsKey("negrita"))
		      buttons.get("negrita").setSelected (StyleConstants.isBold (attrib));
		   if (buttons.containsKey("cursiva"))
		      buttons.get("cursiva").setSelected (StyleConstants.isItalic (attrib));
		   if (buttons.containsKey("subrayado"))
		      buttons.get("subrayado").setSelected (StyleConstants.isUnderline (attrib));
		
		   //No selection and no list being built => finds out type of list and enables correct button
		   if ((iEndSel - iStartSel == 0) && !bBuildingList) {
		      sCurrentList = document.getParagraphElement(textArea.getCaretPosition()).getParentElement().getParentElement().getName();
		      if (buttons.containsKey("insertarVineta")) {
		         if (sCurrentList.equals ("ul"))
		            buttons.get("insertarVineta").setSelected (true);
		         else buttons.get("insertarVineta").setSelected (false);
		      }
		      if (buttons.containsKey("insertarNumeracion")) {
		         if (sCurrentList.equals ("ol"))
		            buttons.get("insertarNumeracion").setSelected (true);
		         else buttons.get("insertarNumeracion").setSelected (false);
		      }
		
		   //If there is a selection, displays all list buttons as disabled   
		   } else {
		   	      if (buttons.containsKey("insertarVineta"))
			         buttons.get("insertarVineta").setSelected (false);
			      if (buttons.containsKey("insertarNumeracion"))
			         buttons.get("insertarNumeracion").setSelected (false);
		   }
		
		   //It the maximum/minimum size is reached, disable the corresponding button 
		   if (buttons.containsKey("aumentar"))
		      buttons.get("aumentar").enable (true);
		   if (buttons.containsKey("disminuir"))
		      buttons.get("disminuir").enable (true);
		   if ((attrib.isDefined (CSS.Attribute.FONT_SIZE)) && (iEndSel - iStartSel == 0)) {
		      if (attrib.getAttribute(CSS.Attribute.FONT_SIZE).toString().equals ("7") &&
			      buttons.containsKey("aumentar"))
		         buttons.get("aumentar").enable (false); 
		      if (attrib.getAttribute(CSS.Attribute.FONT_SIZE).toString().equals ("1") &&
			      buttons.containsKey("disminuir")) 
		         buttons.get("disminuir").enable (false);
		   }  
		
		   //Is the name of the tag within which the current paragraph is "li"? 
		   bInsideList = document.getParagraphElement(iStartSel).getParentElement().getName().equals("li") ||
		                 document.getParagraphElement(iEndSel).getParentElement().getName().equals("li");
		   		   
		   //No change to line spacing if allowed if the caret is positioned inside a list 
		   if (buttons.containsKey("interlineado"))
		      if (bInsideList) {
		         buttons.get("interlineado").enable (false);	
		      } else buttons.get("interlineado").enable (true);
			
		   //No table insertion is allowed if the caret is positioned inside a list
		   if (buttons.containsKey("insertarTabla"))
		      if (bInsideList) {
		         buttons.get("insertarTabla").enable (false);	
		      } else buttons.get("insertarTabla").enable (true);
		
		   //Copying is only allowed provided that text has been selected
		   if (buttons.containsKey("copiarOPegar") && !bCopy) {
		      if (iEndSel - iStartSel != 0)
		         buttons.get("copiarOPegar").enable (true);
		      else buttons.get("copiarOPegar").enable (false);		   
		   }
		}
	}
	
	/**
	 * Updates the status of all toolbar buttons every time the caret moves.
	 */
	public void caretUpdate(CaretEvent e) {
		updateUI (e.getDot());
		refresh.start();		
    }
	
	/**
	 * Purges the keyTyped event if a UNICODE key has been pressed. Used, for example, when selected text 
	 * is made up of different types of blocks (tables and lists).
	 */
	public void keyTyped(KeyEvent e) {
		if (bPurgeKeyPress) {
		   e.consume();
		   char tecla = e.getKeyChar();
		   if ((tecla != '`') && (tecla != '´') && (tecla != '^') && (tecla != '¨'))
		      bPurgeKeyPress = false;
		}	
	}
	
	/**
	 * Every time the window is activated, the text area grabs the focus.
	 */
	public void windowActivated (WindowEvent e) {
		if (bGrabFocus) {
		   textArea.grabFocus();
		   bGrabFocus = false;
		}
	}
	
	/**
	 * Enables the undo button on any text change.
	 */
	public void undoableEditHappened (UndoableEditEvent e) {
		if (buttons.containsKey("deshacer") && (undo != null)) {
		   undo.addEdit (e.getEdit());
		   buttons.get("deshacer").enable (true);	   
		}
	}
	
	/**
	 * Switches the window's favicon to its window-maximised version.
	 */
	public void windowDeiconified (WindowEvent arg0) {
		setIcon (Global.FAVICON_ED);
	}
	
	/**
	 * Switches the window's favicon to its window-minimised version.
	 */
	public void windowIconified(WindowEvent arg0) {
		setIcon (Global.FAVICON_ED_MIN);
	}
		
	/**
	 * Disposes of any window still open (including the editor's) and exits the application.
	 */
	public void windowClosing (WindowEvent e) {
		//Disposes of all secondary windows
		tableConfig.close();
		PBDQkey.close();
	    login.close();
	    	    	
		//Disposes of the editor's window. Since this is the only remaining
	    //window, this triggers the program to exit which, in turn, triggers the
	    //execution of the shutdown hook (document save).
		close();
		System.exit (0);
	}
	
	//Compulsory interface methods
	public void windowClosed(WindowEvent arg0) {}
	public void windowDeactivated(WindowEvent arg0) {}
	public void windowOpened(WindowEvent arg0) {}
	public void mouseDragged(MouseEvent arg0) {}
}
