package editor;

import gui.Canvas;

import java.awt.Dimension;
import java.awt.event.ActionListener;
import java.util.LinkedHashMap;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JPanel;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import utils.Global;

/**
 * Stores in an indexed array all the buttons, visible or not, making up a tool bar. The index allows 
 * filtering out those buttons not visible.
 * @author hqcasanova
 * 
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 */
public class Toolbar {
   private LinkedHashMap<String, Button> visible = new LinkedHashMap<String, Button> (); 
   //index for visible buttons
   private Button[] toolbar;	 //stores tool bar buttons
   private JPanel panel;		 //container for the whole tool bar
   private int iButtonNo;     	 //total number of buttons (visible or not)
   private int iOrientation;     //layout direction
   private boolean bVertical;    //horizontal or vertical tool bar?
   private boolean bDynSpacing;  //dynamic spacing between elements?
        
   /**
    * Customised constructor. Populates a new button array and indexes the visible ones. 
    * @param buttonList Button list retrieved from the configuration XML file.
    * @param vertical Flag for displaying buttons vertically.
    * @param spacing Flag for spreading buttons out on the horizontal/vertical space available.
    */
   public Toolbar (NodeList buttonList, boolean vertical, boolean spacing) {
	  int i;
	  
	  //Initialising
      iButtonNo = buttonList.getLength();
      toolbar = new Button[iButtonNo];
      panel = new JPanel();
      
      //Set orientation
      this.bVertical = vertical;
      if (vertical)
         iOrientation = BoxLayout.Y_AXIS;
      else iOrientation = BoxLayout.X_AXIS;
      panel.setLayout (new BoxLayout (panel, iOrientation));
            
      //Sets dynamic spacing
      this.bDynSpacing = spacing;
      if (spacing)
         setButtonSpace (panel);
      
      //Populates and indexes the button array
      for (i = 0; i < iButtonNo; i++) {
	      toolbar[i] = new Button ((Element) buttonList.item(i));
	      if (toolbar[i].isVisible())
	         visible.put (toolbar[i].getName(), toolbar[i]);
      }
   }
   
   /**
    * Retrieves the panel containing the button tool bar.
    * @return Swing component representing the panel.
    */
   public JPanel getPanel() {
	   return panel;
   }
   
   /**
    * Retrieves the named array with all visible buttons.
    * @return Visible button array.
    */
   public LinkedHashMap<String, Button> getVisibles() {
	  return visible; 
   }
      
   /**
	* Returns whether a given button is visible or not by means of the visible button index.
	* @param name Button name whose visibility is to be determined.
	* @return True if the button is found using the visible button index. False, if not.
	*/
   public boolean isVisible (String name) {
	  Button searched;
		
	  searched = visible.get (name);
	  if (searched != null) 
	     return true;
	  else return false;
   }
   
   /**
    * Sets up the Swing buttons making up the tool bar   
    * @param iIconWidth Width of all icons used for buttons 
    * @param iIconHeight Height of all icons used for buttons
    * @param iIconPadding Space between image and button edges.
    * @param sIconPath Path for button icon files
    * @param sExtension Icon file extension
    * @param sDropDownName Name for the drop-down menu 
    * @param dropDownItems List of items to display in the drop-down menu
    * @param listener Listener class to manage events from buttons.
    */
   public void buttoniseVisibles (int iIconWidth, int iIconHeight, String sIconPath, 
		                          String sExtension, String sDropDownName, String[] dropDownItems,
		                          ActionListener listener) {
	  	  	  
	   for (String nombre : visible.keySet()) {
		   if (!nombre.equals (sDropDownName))
	  	      visible.get(nombre).buttonise (iIconWidth, iIconHeight, sIconPath, sExtension, listener);
	       else visible.get(nombre).buttonise (dropDownItems, listener);
	   }
   }
   
   /**
    * Stretches out the space between toolbar elements.
    * @param panel Container panel of toolbar elements.
    */
   private void setButtonSpace (JPanel panel) {
	   if (bVertical)
	      panel.add (Box.createVerticalGlue());
	   else panel.add (Box.createHorizontalGlue());
   }
     
   /**
    * Groups buttons and renders them inside a titled panel.
    * @param sButtonList List of names of buttons making up the group.
    * @param sTitle Border title. If empty, no border is drawn.
    * @param iSpacing Spacing between buttons in pixels
    * @return Panel with grouped buttons
    */
   public JPanel groupVisibles (String sButtonList, String sTitle, int iSpacing) {
	   JPanel group;		//panel containing group of buttons
	   String[] button;		//list of separate button names to be grouped
	   int i;
	   
	   //Sets up the titled panel
	   group = new JPanel();
	   group.setLayout (new BoxLayout (group, iOrientation));
	   
	   //Adds only those buttons that are visible
	   button = sButtonList.split (Global.TOKEN_SEPARATOR);
	   for (i = 0; i < button.length; i++)
		   if (isVisible (button[i])) {
		      group.add (visible.get(button[i]).getComponent());
		      
		      //Inserts space of a certain width or height between panel elements
		      if (bVertical)
			     group.add (Box.createRigidArea (new Dimension (0, iSpacing)));
			  else group.add (Box.createRigidArea (new Dimension (iSpacing, 0)));		      
		   }
	   	   
	   //If buttons were indeed added...
	   if (group.getComponentCount() != 0) {
		  
          //Surround group of buttons with a title border 		   
	      panel.add (Canvas.titledButtonPanel (group, sTitle, bVertical));
	      
	      //Add flexible space if multiple groups to be spread out
	      if (bDynSpacing)
	         setButtonSpace (panel);
	   } else return null;
	   return group;
   }   
}