package gui;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Rectangle;
import java.awt.Window;

import javax.swing.JFrame;

/**
 * Builds a window from a given canvas. Fine features such as the windows's title 
 * or the close button are configurable.
 * @author hqcasanova
 * 
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 */
public class CanvasWindow extends Canvas {
    private JFrame window;			
        
    /**
     * Customised constructor. Sets the window's and canvas's properties.
     * @param sTitle New dialogue's title
     * @param bgColorTopPanel Background color for the top area. A null value leaves
     * the default color.	 
     * @param bgColorBottomPanel Background color for the bottom area. A null value leaves
     * the default color.	 
     * @param centreColor Background color for the centre area. A null value leaves
     * the default color.
     * @param bChangeDim Is the dialogue resizable?
     * @param exitMode What happens when the X button is pressed?
     * @param sFavicon Windows's favicon path
     * @param bRemoveButtons Remove X and minimise buttons (effective only if using native L&F)
     * @param bRoundedTop True if top panel are to have rounded corners.
     * @param bRoundedBottom True if bottom panel are to have rounded corners. 
     */
	public CanvasWindow (String sTitle, Color bgColorTopPanel, Color bgColorBottomPanel, 
			             Color centreColor, boolean bChangeDim, int exitMode, 
			             String sFavicon, boolean bRemoveButtons,
			             boolean bRoundedTop, boolean bRoundedBottom) {
		
		//Builds the canvas for the window
		super (bgColorTopPanel, bgColorBottomPanel, centreColor, bRoundedTop, bRoundedBottom);
		
		//If title specified, the title bar is kept
		if (sTitle != null) {
		   JFrame.setDefaultLookAndFeelDecorated (true);
		   window = new JFrame (sTitle);
		   
		//No title text given => no title bar
		} else {
			   JFrame.setDefaultLookAndFeelDecorated (false); 
			   window = new JFrame ();
			   window.setUndecorated (true);
		}
		
		//Sets rest of properties: exit mode, resizability, favicon...
		window.setDefaultCloseOperation (exitMode);
		window.setResizable (bChangeDim);
		if (bRemoveButtons)
		   removeButtons (window);
		setWindow (window);
		setCanvas();
		setIcon(sFavicon);
		
		//Prepares busy mouse pointer: links wait cursor with glass pane
		window.getGlassPane().setCursor (Cursor.getPredefinedCursor (Cursor.WAIT_CURSOR));		
	}	
	
	/**
	 * Customised constructor for windows without rounded panels.	 
	 */
	public CanvasWindow (String sTitle, Color colorSup, Color colorInf, Color centreColor, boolean bChangeDim, 
	        			 int exitMode, String sFavicon, boolean bRemoveButtons) {
		this (sTitle, colorSup, colorInf, centreColor, bChangeDim, 
	          exitMode, sFavicon, bRemoveButtons, false, false);
	}
	
	/**
	 * Retrieves the windows's root container.
	 * @return JFrame object.
	 */
	public JFrame getWindow () {
		return window;
	}
	
	/**
	 * Sets the canvas for the present window.
	 */
	private void setCanvas () {
		window.setContentPane (getCanvas());
	}
		
	/**
	 * Removes the windows's canvas. The end result is an empty window that can
	 * be used for further manipulation.
	 */
	protected void removeCanvas () {
		window.getContentPane().remove (getCanvas());
	}
	
	/**
	 * Allows changing the action taking place when the "X" button is pressed.
	 * @param iExitMode New exit mode. JFrame.EXIT_ON_CLOSE constant supported.
	 */
	protected void changeExitMode (int iModoSalida) {
		window.setDefaultCloseOperation (iModoSalida);
	}
	
	/**
	 * Changes the mouse pointer to "busy" mode for user feedback.   
	 * @param bActivity True if there is some kind of activity. False if not.
	 */
    public void busy (boolean bActividad) {
		window.getGlassPane().setVisible (bActividad);
	}
    
    /**
     * Maximises window and reports it to the system 
     */
    protected void maximise() {
    	Rectangle workArea = getWorkArea();
    	window.setPreferredSize (workArea.getSize());
		window.setMaximizedBounds (workArea);
		window.setExtendedState (window.getExtendedState() | JFrame.MAXIMIZED_BOTH);
    }
}